// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

// NOTE: Please read README.txt before browsing this code.

// include library header file.
#include <SpatialIndex.h>

using namespace SpatialIndex;
using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2

long computeCapacity(long p, long f)
{
	return (long)ceil((double)(p*(f-1)+60*f-12)/(double)(56*f-32));		

}

int main(int argc, char** argv)
{
	try
	{
		if (argc != 6)
		{
			cerr << "Usage: " << argv[0] << " database_file tree_file pagesize embedding_fanout insert_file." << endl;
			return -1;
		}
		long capacity = computeCapacity(atoi(argv[3]), atoi(argv[4]));
		cerr << "capacity :"<<capacity<<endl;



		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open database file " << argv[1] << "." << endl;
			return -1;
		}

		ifstream fin1(argv[5]);
		if (! fin1)
		{
			cerr << "Cannot open insertion file " << argv[4] << "." << endl;
			return -1;
		}

		// Create a new storage manager with the provided base name and a 4K page size.
		string baseName = argv[2];
		IStorageManager* diskfile = StorageManager::createNewDiskStorageManager(baseName, 1024);
		StorageManager::IBuffer* file = StorageManager::createNewRandomEvictionsBuffer(*diskfile, 10, false);
			// applies a main memory random buffer on top of the persistent storage manager
			// (LRU buffer, etc can be created the same way).

		// Create a new, empty, RTree with dimensionality 2, minimum load 70%, using "file" as
		// the StorageManager and the RSTAR splitting policy.
		long indexIdentifier;
		ISpatialIndex* tree = BTree::createNewBTree(*file, 0.7, capacity, capacity ,1, atoi(argv[4]), SpatialIndex::BTree::BV_PLUS, indexIdentifier);

		unsigned long count = 0;
		unsigned long indexIO = 0;
		unsigned long leafIO = 0;
		unsigned long id, op;
		double x[1];
		double y[1];
		Tools::ResourceUsage ru;

		tree->stopCountingAffectedNodes();
		while (fin)
		{
			fin >> op >> id >> x[0] >> y[0];
			if (! fin.good()) continue; // skip newlines, etc.

			if (op == INSERT)
			{
				Point p = Point(x, 1);
				tree->insertData(0, 0, p, id);
			}
		}
		tree->clearCountAffectedNodes();
		tree->startCountingAffectedNodes();
		long insertcount=0;
		ru.start();
		while (fin1)
		{
			fin1 >> op >> id >> x[0] >> y[0];
			if (! fin1.good()) continue; // skip newlines, etc.

			if (op == INSERT)
			{
				Point p = Point(x, 1);
				tree->insertData(0, 0, p, id);
				insertcount++;
			}
		}
		ru.stop();	
		Tools::ResourceUsage hashru;
		hashru.start();
		tree->computeHashes();
		hashru.stop();

		cerr << "Operations: " << count << endl;
		cerr << *tree;
		cerr << "Index I/O: " << indexIO << endl;
		cerr << "Leaf I/O: " << leafIO << endl;
		cerr << "Buffer hits: " << file->getHits() << endl;
		cerr << "Insert Seconds: " << ru.getTotalTime() << endl;
		cerr << "Hash Seconds: " << hashru.getTotalTime() << endl;
		cerr << "insert: "<<insertcount<<endl;

		delete tree;
		delete file;
		delete diskfile;
			// delete the buffer first, then the storage manager
			// (otherwise the the buffer will fail writting the dirty entries).
	}
	catch (Tools::Exception& e)
	{
		cerr << "******ERROR******" << endl;
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "******ERROR******" << endl;
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
