// Spatial Index Library
//
// Copyright (C) 2003 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <SpatialIndex.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>


using namespace SpatialIndex;

NwayTree::NwayTree():
	m_root(0),
	m_rootID(-1),
	m_nodeID(-1),
	m_fanout(0),
	m_treeDigest(0)
{
}

NwayTree::NwayTree(long fanout):
	m_root(0),
	m_rootID(-1),
	m_nodeID(-1),
	m_fanout(fanout),
	m_treeDigest(0)
{
}

NwayTree::NwayTree(long fanout, byte* const data):
	m_root(0),
	m_rootID(-1),
	m_nodeID(-1),
	m_fanout(fanout),
	m_treeDigest(0)
{
	loadFromByteArray(data);
}

	
NwayTree::~NwayTree()
{
	if (m_root!=0)
		delete m_root;
}

//
// IObject interface
//
NwayTree* NwayTree::clone() throw (Tools::NotSupportedException)
{
	return new NwayTree();
}

//
// ISerializable interface
//
unsigned long NwayTree::getByteArraySize()
{
	unsigned long len=sizeof(long)+sizeof(long)+sizeof(long)+DIGEST_LENGTH;
	return len;
}

unsigned long NwayTree::loadFromByteArray(byte* const data)
{
	byte* ptr = data;

	memcpy(&m_rootID, ptr, sizeof(long));
	ptr+=sizeof(NwayNode*);

	memcpy(&m_nodeID, ptr, sizeof(long));
	ptr+=sizeof(long);

	memcpy(&m_fanout, ptr, sizeof(long));
	ptr+=sizeof(long);

	
	m_treeDigest= new byte[DIGEST_LENGTH];
	memcpy(m_treeDigest, ptr, DIGEST_LENGTH);
	ptr+=DIGEST_LENGTH;
	
	return getByteArraySize();
}

void NwayTree::storeToByteArray(unsigned long& len, byte** data)
{
	len = getByteArraySize();
	*data = new byte[len];
	byte* ptr = *data;
	
	memcpy(ptr, &m_rootID, sizeof(long));
	ptr+=sizeof(long);
	
	memcpy(ptr, &m_nodeID, sizeof(long));
	ptr+=sizeof(long);

	memcpy(ptr, &m_fanout, sizeof(long));
	ptr+=sizeof(long);
	
	memcpy(ptr, m_treeDigest, DIGEST_LENGTH);
	ptr+=DIGEST_LENGTH;
	
	assert(len==(ptr-*data));
}

void NwayTree::bulkloadTree(long* keys, long keysize, byte** indigests)
{
	long i, j, k;
	std::vector<NwayNode*> nodes;
	k=0;
	long* nodeKeys=new long[m_fanout];
	NwayNode** nodeChild=0;
	byte** digests;
	long* nodeChildID=0;
	long level=0;
	long children=0;
	if (indigests==0)
		digests=0;
	else
		digests= new byte*[m_fanout];	
	m_cputime=0.0;
	m_hashComputed=0;
	for (i=0;i<keysize;i++)
	{
		j=i%m_fanout;
		if (j!=0 || i==0)
			children++;
		if (j==0 && i!=0)
		{
			NwayNode* node= new NwayNode(this, level, ++m_nodeID,children, nodeKeys, nodeChild, nodeChildID, digests);
			nodes.push_back(node);
			children=1;
		}
		nodeKeys[j]=keys[i];
		if (indigests!=0)
		{
			digests[j]= new byte[DIGEST_LENGTH];
			memcpy(digests[j], indigests[i], DIGEST_LENGTH);
		}
	}
	NwayNode* node= new NwayNode(this, level, ++m_nodeID,children, nodeKeys, nodeChild, nodeChildID, digests);
	nodes.push_back(node);
	children=0;
	nodeChild= new NwayNode*[m_fanout];
	nodeChildID = new long[m_fanout];
	if (digests==0)
		digests=new byte*[m_fanout];
	while(nodes.size()>1)
	{
		long size=nodes.size();
		level++;
		children=0;
		for (i=0; i<size;i++)
		{
			NwayNode* node= (NwayNode*)nodes.front();
			nodes.erase(nodes.begin());
			j=i%m_fanout;
			if (j!=0 || i==0)
				children++;
			if (j==0 && i!=0)
			{
				NwayNode* node= new NwayNode(this, level, ++m_nodeID,children, nodeKeys, nodeChild, nodeChildID, digests);
				nodes.push_back(node);
				children=1;
			}
			nodeKeys[j]=node->m_pKey[0];
			nodeChild[j]=node;	
			nodeChildID[j]=node->m_identifier;
			digests[j]=new byte[DIGEST_LENGTH];
			node->computeNodeDigests(digests[j]);
		}
		NwayNode* node= new NwayNode(this, level, ++m_nodeID,children, nodeKeys, nodeChild, nodeChildID, digests);
		nodes.push_back(node);
	}
	m_root=(NwayNode*)nodes.front();
	nodes.erase(nodes.begin());
	m_rootID=m_root->m_identifier;
	if(m_treeDigest==0)
		m_treeDigest=new byte[DIGEST_LENGTH];
	m_root->computeNodeDigests(m_treeDigest);
	delete[] nodeChild;
	delete[] digests;
	delete[] nodeChildID;
	delete[] nodeKeys;
}

void NwayTree::writeNode(NwayNode* node, byte** buffer, unsigned long& dataLength)
{
	node->storeToByteArray(dataLength, buffer);	
}

void NwayTree::writeTree(byte** buffer, unsigned long& dataLength)
{
	std::vector<NwayNode*> nodes;
	std::vector<byte*> datas;
	std::vector<unsigned long> datalen;
	byte* tmp;
	unsigned long tmpLen=0;
	dataLength=0;
	if (m_root==0)
		return;
	this->storeToByteArray(tmpLen, &tmp);
	datas.push_back(tmp);
	datalen.push_back(tmpLen);	
	dataLength+=tmpLen;	
	nodes.push_back(m_root);
	while(nodes.size()>0)
	{
		NwayNode* node=(NwayNode*)nodes.front();
		nodes.erase(nodes.begin());
		writeNode(node, &tmp, tmpLen);
		datas.push_back(tmp);
		datalen.push_back(tmpLen);
		dataLength+=tmpLen;
		if (node->m_level!=0)
		{
			for (long i=0; i< node->m_children; i++)
				nodes.push_back(node->m_pChild[i]);
		}
	}
	*buffer= new byte[dataLength];
	byte* ptr = *buffer;
	long size=datas.size();
	for (long i=0; i<size; i++)
	{
		tmp=(byte*)datas.front();
		datas.erase(datas.begin());
		tmpLen=(long)datalen.front();
		datalen.erase(datalen.begin());
		memcpy(ptr, tmp, tmpLen);
		ptr+=tmpLen;
		delete[] tmp;
	}
}

NwayNode* NwayTree::readNode(byte*& buffer)
{
	NwayNode* n=new NwayNode(this);
	unsigned long len=n->loadFromByteArray(buffer);
	buffer+=len; 
	return n;
}

void NwayTree::readTree(byte* buffer)
{
	byte* tmp=buffer;
	std::vector<NwayNode*> nodes;
	unsigned long len=this->loadFromByteArray(buffer);
	buffer+=len;
	long i;
	for (i=0; i<=m_nodeID;i++)
	{
		//NwayNode* node= new NwayNode(tree, tmp);
		NwayNode* node = readNode(buffer);
		nodes.push_back(node);
	}	
	long size=nodes.size();	
	for (i=0; i<size; i++)
	{
		NwayNode* node=(NwayNode*)nodes.front();
		nodes.erase(nodes.begin());
		long set=0;
		for (long j=0; j<nodes.size();j++)
		{
			NwayNode* tn = (NwayNode*)nodes[j];
			for (long k=0; k<node->m_children; k++)
			{
				if (tn->m_identifier==node->m_pChildID[k])
				{
					node->m_pChild[k]=tn;
					set++;	
					break;
				}
			}
			if (set==node->m_children)
				break;
		}
		if (node->m_identifier==m_rootID)
			m_root=node;
	}
	buffer=tmp;
}

void NwayTree::getLeafDigests(byte** digests)
{
	std::vector<NwayNode*> leaf;
	std::vector<NwayNode*> index;
	NwayNode* tn=m_root;
	if (tn->m_level==0)
		leaf.push_back(tn);
	else
		index.push_back(tn);
	long i;
	while(index.size()>0)
	{
		tn=index.front();
		index.erase(index.begin());
		for(i=0;i<tn->m_children;i++)
		{	if (tn->m_pChild[i]->m_level==0)
				leaf.push_back(tn->m_pChild[i]); 
			else
				index.push_back(tn->m_pChild[i]);
		}
	}
	long k=0;
	while(leaf.size()>0)
	{
		tn=leaf.front();
		leaf.erase(leaf.begin());
		for (i=0;i<tn->m_children;i++)
			memcpy(digests[k++], tn->m_pDigest[i],DIGEST_LENGTH);
	}
}

void NwayTree::constructVOs(long searchKey, byte**& voDigest, long& voSize, long** pos, long& posSize)
{
	NwayNode* tn=m_root;
	std::vector<byte*> vos;
	std::vector<long> positions;
	long i, slot;
	while(true)
	{
		slot=tn->getSlot(searchKey);
		if (tn->m_children>1)
			positions.push_back(slot);
		for (i=0;i<tn->m_children;i++)
		{
			if (i!=slot)	
				vos.push_back(tn->m_pDigest[i]);			
		} 
		if (tn->m_level==0)
			break;
		tn=tn->m_pChild[slot];
	}
	voSize=vos.size();
	voDigest= new byte*[voSize];
	posSize=positions.size();
	*pos=(long*) malloc(posSize*sizeof(long));
	i=voSize-1;
	for (long j=0;j<voSize;j++)
	{
		byte* tmp=(byte*)vos[j];
		voDigest[i]=tmp;
		i--;
	}
	i=posSize-1;
	for (long j=0;j<posSize;j++)
	{
		long t=(long)positions[j];
		(*pos)[i]=t;
		i--;
	}

	vos.clear();
	positions.clear();
}

void NwayTree::showTree()
{
	std::vector<NwayNode*> nodes;
	NwayNode* tmp;
	if (m_root==0)
		return;
	nodes.push_back(m_root);
	while(!nodes.empty())
	{
		tmp=nodes[0];
		nodes.erase(nodes.begin());
		tmp->showNode();
		if (tmp->m_level!=0)
		{
			for(long i=0; i<tmp->m_children;i++)
				nodes.push_back(tmp->m_pChild[i]);
		}
	}
	printf("\n root ID: %d tree level: %d tree digest: ", m_rootID, m_root->m_level+1);
	unsigned long j=0;
	for (j=0; j<DIGEST_LENGTH; j++)
		printf("%02x",m_treeDigest[j]); 
	printf("\n");
}

void NwayTree::storeToFile(char* file)
{
	int treeFile;
	int cMode = O_CREAT | O_RDWR | O_TRUNC;
	treeFile = open(file, cMode, 0644);
	ssize_t byteswritten;	
	byte* buffer;
	unsigned long length;
	writeTree(&buffer, length);
	//std::cout<<"tree length "<<length<<std::endl;
	byteswritten=write(treeFile, buffer, length);
	//fsync(treeFile);
}

void NwayTree::loadFromFile(char* file)
{
	int treeFile;
	int cMode = O_RDWR;
	treeFile = open(file, cMode, 0644);
	ssize_t bytesread;
	unsigned long length;
	off_t seeke = lseek(treeFile, 0, SEEK_END);
	lseek(treeFile, -1*seeke, SEEK_END);
	length = (unsigned long)seeke;
	byte* buffer= new byte[length];
	bytesread=read(treeFile,buffer, length); 
	readTree(buffer);
}

std::ostream& SpatialIndex::operator<<(std::ostream& os, const NwayTree& pt)
{
	return os;
}
