// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

// NOTE: Please read README.txt before browsing this code.

// include library header file.
#include <SpatialIndex.h>

using namespace SpatialIndex;
using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2

// example of a Visitor pattern.
// see RTreeQuery for a more elaborate example.
class MyVisitor : public IVisitor
{
public:
	unsigned long m_indexIO;
	unsigned long m_leafIO;

public:
	MyVisitor() : m_indexIO(0), m_leafIO(0) {}

	void visitNode(const INode& n) {
		if (n.isLeaf()) m_leafIO++;
		else m_indexIO++;
	}



	void visitData(const IData& d)
	{
		IShape* pS;
		d.getShape(&pS);
			// do something.
		Point key;
		pS->getCenter(key);	
		delete pS;
		cout<< "Query results: pid key "<<endl;
		cout << d.getIdentifier() << " "<< key.m_pCoords[0]<<endl;
			// the ID of this data entry is an answer to the query. I will just print it to stdout.
	}

	void visitData(std::vector<const IData*>& v) {}
	
	void updateVOs(byte** digest, long num, long pos) { delete[] digest;}
	void updateVOs(byte** digest, long num, long* pos, long posnum) { delete[] digest;}

};


long computeCapacity(long p)
{
	return (long)ceil((double)(p-24)/56.0);		

}

int main(int argc, char** argv)
{
	try
	{
		if (argc != 6)
		{
			cerr << "Usage: " << argv[0] << " input_file tree_file pagesize embedding_fanout query_type [pointsearch | rangequery]." << endl;
			return -1;
		}

		int queryType = 0;
		long capacity = computeCapacity(atoi(argv[3]));
		cerr << "capacity :"<<capacity<<endl;

		if (strcmp(argv[5], "pointsearch") == 0) queryType = 0;
		else if (strcmp(argv[5], "rangequery") == 0) queryType = 1;
		else
		{
			cerr << "Unknown query type." << endl;
			return -1;
		}

		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open data file " << argv[1] << "." << endl;
			return -1;
		}

		// Create a new storage manager with the provided base name and a 4K page size.
		string baseName = argv[2];
		IStorageManager* diskfile = StorageManager::createNewDiskStorageManager(baseName, atoi(argv[3]));
		StorageManager::IBuffer* file = StorageManager::createNewRandomEvictionsBuffer(*diskfile, 10, false);
			// applies a main memory random buffer on top of the persistent storage manager
			// (LRU buffer, etc can be created the same way).

		// Create a new, empty, RTree with dimensionality 2, minimum load 70%, using "file" as
		// the StorageManager and the RSTAR splitting policy.
		long indexIdentifier;
		ISpatialIndex* tree = BTree::createNewBTree(*file, 0.7, capacity, capacity ,1, atoi(argv[4]), SpatialIndex::BTree::BV_PLUS, indexIdentifier);

		unsigned long count = 0;
		unsigned long indexIO = 0;
		unsigned long leafIO = 0;
		unsigned long id, op;
		double x[1];
		double y[1];
		Tools::ResourceUsage ru;

		ru.start();

		while (fin)
		{
			fin >> op >> id >> x[0] >> y[0];
			if (! fin.good()) continue; // skip newlines, etc.

			if (op == INSERT)
			{
				Point p = Point(x, 1);

				//ostringstream os;
				//os << r;
				//string data = os.str();
					// associate some data with this region. I will use a string that represents the
					// region itself, as an example.
					// NOTE: It is not necessary to associate any data here. A null pointer can be used. In that
					// case you should store the data externally. The index will provide the data IDs of
					// the answers to any query, which can be used to access the actual data from the external
					// storage (e.g. a hash table or a database table, etc.).
					// Storing the data in the index is convinient and in case a clustered storage manager is
					// provided (one that stores any node in consecutive pages) performance will improve substantially,
					// since disk accesses will be mostly sequential. On the other hand, the index will need to
					// manipulate the data, resulting in larger overhead. If you use a main memory storage manager,
					// storing the data externally is highly recommended (clustering has no effect).
					// A clustered storage manager is NOT provided yet.
					// Also you will have to take care of converting you data to and from binary format, since only
					// array of bytes can be inserted in the index (see RTree::Node::load and RTree::Node::store for
					// an example of how to do that).

				//tree->insertData(data.size() + 1, reinterpret_cast<const byte*>(data.c_str()), r, id);
				//cout<<"Insert "<<p.m_pCoords[0]<<endl;
				tree->insertData(0, 0, p, id);
					// example of passing zero size and a null pointer as the associated data.
			}
			else if (op == DELETE)
			{
				Point p= Point(x,1);
				cout<<"delete "<<p.m_pCoords[0]<<endl;
				if (tree->deleteData(p, id) == false)
				{
					cerr << "******ERROR******" << endl;
					cerr << "Cannot delete id: " << id << " , count: " << count << endl;
					return -1;
				}
			}
			else if (op == QUERY)
			{

				MyVisitor vis;
				cout << "query "<<id<<" "<<x[0]<<" "<<y[0]<<endl;

				if (queryType == 0)
				{
					Point p = Point(x, 1);
					tree->pointLocationQuery(p, vis);
				}
				else if (queryType == 1)
				{
					LineSegment r = LineSegment(x, y, 1);
					tree->containsWhatQuery(r, vis);
				}
				indexIO += vis.m_indexIO;
				leafIO += vis.m_leafIO;

				
			}

			if (((count+1) % 10000) == 0 && count!=0)
				cerr << count+1 << endl;
			count++;
		}

		tree->computeHashes();
		ru.stop();
		//tree->showTree();
		tree->computeFanout();

		cerr << "Operations: " << count << endl;
		cerr << *tree;
		cerr << "Index I/O: " << indexIO << endl;
		cerr << "Leaf I/O: " << leafIO << endl;
		cerr << "Buffer hits: " << file->getHits() << endl;
		cerr << "Seconds: " << ru.getTotalTime() << endl;
		cerr << "Index ID: " << indexIdentifier << endl;
		

		bool ret = tree->isIndexValid();
		if (ret == false) cerr << "ERROR: Structure is invalid!" << endl;
		else cerr << "The stucture seems O.K." << endl;

		delete tree;
		delete file;
		delete diskfile;
			// delete the buffer first, then the storage manager
			// (otherwise the the buffer will fail trying to write the dirty entries).
	}
	catch (Tools::Exception& e)
	{
		cerr << "******ERROR******" << endl;
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "******ERROR******" << endl;
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
