// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include "../spatialindex/SpatialIndexImpl.h"

#include "BTree.h"
#include "Node.h"
#include "Index.h"

using std::stack;
using std::vector;
using namespace std;
using namespace SpatialIndex::BTree;

//
// Tools::IObject interface
//
Tools::IObject* Node::clone() throw (Tools::NotSupportedException)
{
	throw Tools::NotSupportedException("IObject::clone should never be called.");
}

//
// Tools::ISerializable interface
//
unsigned long Node::getByteArraySize()
{
	unsigned long len=	
		(sizeof(long) +
		sizeof(long) +
		sizeof(long) +
		sizeof(unsigned long) +
		sizeof(long) +
		sizeof(long) +
		(m_children * ( sizeof(long)  + sizeof(long) + 2*sizeof(int)) ) +
		+ m_embeddingLength ); 
	if (m_rootID!=-1)
		len+= DIGEST_LENGTH;
	return len;
}

unsigned long Node::loadFromByteArray(byte* const data)
{
	byte* ptr = data;
	long tmp=0;

	// skip the node type information, it is not needed.
	ptr += sizeof(long);

	memcpy(&m_level, ptr, sizeof(long));
	ptr += sizeof(long);

	memcpy(&m_children, ptr, sizeof(long));
	ptr += sizeof(long);

	memcpy(&m_embeddingLength, ptr, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	
	memcpy(&(m_pFirstNode), ptr, sizeof(long));
	ptr += sizeof(long);

	memcpy(&(m_rootID), ptr, sizeof(long));
	ptr += sizeof(long);

	if (m_level>0 && m_embeddingLength>0)
	{
		m_pFirstDigest =  new byte[DIGEST_LENGTH]; 
		memcpy(m_pFirstDigest, ptr, DIGEST_LENGTH);
		ptr += DIGEST_LENGTH;
	}
	else
		m_pFirstDigest=0;

	unsigned long cChild=0;
	for (cChild = 0; cChild < m_children; cChild++)
	{
		memcpy(&(m_pIdentifier[cChild]), ptr, sizeof(long));
		ptr += sizeof(long);

		m_ptrKey[cChild]=m_pTree->m_pointPool.acquire();
		*(m_ptrKey[cChild]) = m_pTree->m_infinitePoint;
		memcpy(&tmp, ptr, sizeof(long));
		ptr +=  sizeof(long);
		m_ptrKey[cChild]->m_pCoords[0]=(double)tmp;
		
		for (int i=0;i<2;i++)
		{
			memcpy(&(m_pID[cChild][i]), ptr, sizeof(int));
			ptr +=  sizeof(int);
		}

		if (m_embeddingLength>0)
		{
			m_pDigest[cChild] = new byte[DIGEST_LENGTH];
			memcpy(m_pDigest[cChild], ptr, DIGEST_LENGTH);
			ptr += DIGEST_LENGTH;
			for (int i=0;i<2;i++)
			{
				if (m_pID[cChild][i]!=-1)
				{	
					m_pHash[cChild][i] = new byte[DIGEST_LENGTH];
					memcpy(m_pHash[cChild][i], ptr, DIGEST_LENGTH);
					ptr +=  DIGEST_LENGTH;
				}
				else
					m_pHash[cChild][i] = 0;
			}
		}
		else
		{
			m_pDigest[cChild]=0;
			m_pHash[cChild][0]=0;
			m_pHash[cChild][1]=0;
		}
	}
	
	if (m_rootID!=-1)	
	{
		restoreEmbeddingTree();
		m_embeddingTree->m_treeDigest=new byte[DIGEST_LENGTH];
		memcpy(m_embeddingTree->m_treeDigest, ptr, DIGEST_LENGTH);
		ptr +=  DIGEST_LENGTH;
	}
	else
		m_embeddingTree=0;
	return getByteArraySize();
}

void Node::storeToByteArray(unsigned long& len, byte** data)
{

	len = getByteArraySize();

	*data = new byte[len];
	byte* ptr = *data;

	long type;
	long tmp;
	
	if (m_level == 0) type = PersistentLeaf;
	else type = PersistentIndex;

	memcpy(ptr, &type, sizeof(long));
	ptr += sizeof(long);

	memcpy(ptr, &m_level, sizeof(long));
	ptr += sizeof(long);

	memcpy(ptr, &m_children, sizeof(long));
	ptr += sizeof(long);
	
	memcpy(ptr, &m_embeddingLength, sizeof(unsigned long));
	ptr += sizeof(unsigned long);

	memcpy(ptr, &(m_pFirstNode), sizeof(long));
	ptr += sizeof(long);

	memcpy(ptr, &(m_rootID), sizeof(long));
	ptr += sizeof(long);


	if (m_embeddingLength!= 0 && m_level!=0)
	{
		memcpy(ptr, m_pFirstDigest, DIGEST_LENGTH);
		ptr += DIGEST_LENGTH;
	}


	unsigned long cChild=0;
	for (cChild = 0; cChild < m_children; cChild++)
	{
		memcpy(ptr, &(m_pIdentifier[cChild]), sizeof(long));
		ptr += sizeof(long);

		tmp=(long)m_ptrKey[cChild]->m_pCoords[0];
		memcpy(ptr, &tmp, sizeof(long));
		ptr += sizeof(long);

		for (int i=0;i<2;i++)
		{
			memcpy(ptr, &(m_pID[cChild][i]), sizeof(int));
			ptr +=  sizeof(int);
		}

		if (m_embeddingLength> 0)
		{
			memcpy(ptr, m_pDigest[cChild], DIGEST_LENGTH);
			ptr += DIGEST_LENGTH;
			for (int i=0;i<2;i++)
			{
				if (m_pID[cChild][i]!=-1)
				{
					memcpy(ptr,m_pHash[cChild][i], DIGEST_LENGTH);
					ptr +=  DIGEST_LENGTH;
				}
			}
		}
	}

	//if (m_embeddingLength > 0 && m_embeddingTree!=0)
	if (m_rootID!=-1)
	{
		memcpy(ptr, m_embeddingTree->m_treeDigest, DIGEST_LENGTH);
		ptr += DIGEST_LENGTH;
	}

	
	assert(len == (ptr - *data) );
}

void Node::restoreEmbeddingTree()
{
	m_embeddingTree=new RBTree();
	m_embeddingTree->m_root=restoreRBNode(m_rootID);
	//m_embeddingTree->m_rootID=m_embeddingTree->m_root->m_id;		
}

RBNode* Node::restoreRBNode(int pos)
{
	if (pos==-1)
		return NULL;
	if (pos==m_children)
	{
		RBNode* node=new RBNode(m_embeddingTree, (long)(m_ptrKey[0]->m_pCoords[0]-1), m_pFirstNode, pos, m_pFirstDigest);
		node->m_leftID=-1;
		node->lDigest=NULL;
		node->m_rightID=-1;
		node->rDigest=NULL;
		return node;
	}
	RBNode* node=new RBNode(m_embeddingTree, (long)(m_ptrKey[pos]->m_pCoords[0]), m_pIdentifier[pos], pos, m_pDigest[pos]);
	node->left=restoreRBNode(m_pID[pos][0]);
	if (node->left==NULL)
	{
		node->m_leftID=-1;
		node->lDigest=NULL;
	}
	else
	{
		node->lDigest=(byte*)malloc(DIGEST_LENGTH);
		memcpy(node->lDigest, m_pHash[pos][0], DIGEST_LENGTH);
		node->m_leftID=node->left->m_id;
	}
	node->right=restoreRBNode(m_pID[pos][1]);
	if (node->right==NULL)
	{
		node->m_rightID=-1;
		node->rDigest=NULL;
	}
	else
	{
		node->rDigest=(byte*)malloc(DIGEST_LENGTH);
		memcpy(node->rDigest, m_pHash[pos][1], DIGEST_LENGTH);
		node->m_rightID=node->right->m_id;
	}
	return node;
}

void Node::updateEmbeddingTree()
{
	buildEmbeddings();	
	long pos=0;
	if (isIndex())
		pos=-1;
	m_embeddingLength=0; //embedding tree's digest anyway will be stored!
	inorderTraverse(m_embeddingTree->m_root, pos);
}

//
// SpatialIndex::IEntry interface
//
long Node::getIdentifier() const
{
	return m_identifier;
}

void Node::getShape(IShape** out) const throw (std::exception)
{
	//*out = m_pTree->readNode(m_pFirstNode).get();

}

//
// SpatialIndex::INode interface
//
unsigned long Node::getChildrenCount() const
{
	return m_children;
}

long Node::getChildIdentifier(unsigned long index) const throw (Tools::IndexOutOfBoundsException)
{
	if (index < -1 || index >= (m_children)) throw Tools::IndexOutOfBoundsException(index);
	
	if (index>=0)
		return m_pIdentifier[index];
	return m_pFirstNode;
}

void Node::getChildShape(unsigned long index, IShape** out) const throw (Tools::IndexOutOfBoundsException, std::exception)
{
	if (index < 0 || index >= m_children) throw Tools::IndexOutOfBoundsException(index);

	*out = new Point(*(m_ptrKey[index]));
}

unsigned long Node::getLevel() const
{
	return m_level;
}

bool Node::isLeaf() const
{
	return (m_level == 0);
}

bool Node::isIndex() const
{
	return (m_level != 0);
}

//
// Internal
//

Node::Node() :
	m_pTree(0),
	m_level(0),
	m_identifier(-1),
	m_children(0),
	m_capacity(0),
	m_ptrKey(0),
	m_pIdentifier(0),
	m_pFirstNode(-1),
	m_pDigest(0),
	m_pFirstDigest(0),
	m_embeddingLength(0),
	m_embeddingTree(0),
	m_pHash(0),
	m_pID(0),
	m_rootID(-1)
{
}

Node::Node(BTree* pTree, long id, unsigned long level, unsigned long capacity) :
	m_pTree(pTree),
	m_level(level),
	m_identifier(id),
	m_children(0),
	m_capacity(capacity),
	m_ptrKey(0),
	m_pIdentifier(0),
	m_pFirstNode(-1),
	m_pDigest(0),
	m_pFirstDigest(0),
	m_embeddingLength(0),
	m_embeddingTree(0),
	m_pHash(0),
	m_pID(0),
	m_rootID(-1)
{

	try
	{
		m_ptrKey = new PointPtr[m_capacity];
		m_pIdentifier = new long[m_capacity];
		m_pDigest = new byte*[m_capacity];
		memset(m_pDigest,0,sizeof(byte*)*m_capacity);		
		
		m_pID = (int**)malloc(m_capacity*sizeof(int*));
		m_pHash = (byte***)malloc(m_capacity*sizeof(byte**));
		for (long i=0; i<m_capacity;i++)
		{
			m_pID[i]=(int*)malloc(2*sizeof(int));
			m_pHash[i]=(byte**)malloc(2*sizeof(byte*));
			for (int a=0;a<2;a++)
			{
				m_pHash[i][a]=0;		
				m_pID[i][a]=-1;
			}
		}
	}
	catch (...)
	{
		delete[] m_ptrKey;
		delete[] m_pIdentifier;
		throw;
	}
}

Node::~Node()
{
	for (unsigned long cChild = 0; cChild < m_children; cChild++)
	{
		if (m_pDigest[cChild] != 0) delete[] m_pDigest[cChild];
		for (int i=0;i<2;i++)
		{
			if (m_pHash[cChild][i]!=0) delete[] m_pHash[cChild][i];
		}
		delete[] m_pID[cChild];
		delete[] m_pHash[cChild];
	}

	delete[] m_pDigest;
	delete[] m_pHash;

	if (m_pFirstDigest!=0)
		delete[] m_pFirstDigest;

	delete[] m_ptrKey;
	delete[] m_pIdentifier;
	delete[] m_pID;
	if (m_embeddingTree!=0)
		delete m_embeddingTree;	
}

Node& Node::operator=(const Node& n)
{
	throw Tools::IllegalStateException("operator =: This should never be called.");
}


//deleteEntry is never called, deletetion is done through remove entry
void Node::deleteEntry(unsigned long index)
{
}

bool Node::isFull()
{
	bool full=false;
	if (m_children==m_capacity)
		full=true;
	return full;
}

bool Node::isFewest()
{
	bool fewest=false;
	if (m_level==m_pTree->m_stats.m_treeHeight-1)
	{
		if (m_children==1)
			fewest=true;
	}
	else if (m_children<=m_minFanout)
		fewest=true;

	return fewest;
}

long Node::getSlot(long searchKey)
{
	long slot, entries;
	
	entries=m_children;
	if (entries==0)
	 	slot=-1;
	else	slot=findKey(searchKey, 0, entries);
	return slot;
}

long Node::findKey(long searchKey, long lo, long hi)
{
	long mid;
	long findslot;
	
	if (hi==lo)
		findslot=bestMatch(searchKey, lo);
	else
	{
		mid = (lo + hi) >> 1;
		switch (findslot = bestMatch(searchKey, mid))
		{
			case -2:
				findslot = findKey(searchKey, lo, mid-1);
				break;
			case -3: 
				findslot= findKey(searchKey, mid+1, hi);
				break;
			default:
				break;
		}
	}
	return findslot;
}

long Node::bestMatch(long searchKey,long slot)
{
	long diff, comp;
	long findslot;

	diff=this->m_pTree->compareKeys(searchKey, (long)m_ptrKey[slot]->m_pCoords[0]);
   	if (diff < 0)
	{
		if (slot==0)
			findslot=-1;
		else if ((comp=this->m_pTree->compareKeys(searchKey, (long)m_ptrKey[slot-1]->m_pCoords[0]))>=0)
			findslot=slot-1;
		else
			findslot= -2;
	}
	else
	{
		if (slot==m_children-1)
			findslot=slot;
		else if ((comp=this->m_pTree->compareKeys(searchKey, (long)m_ptrKey[slot+1]->m_pCoords[0]))<0)
			findslot=slot;
		else if (comp==0)
			findslot=slot+1;
		else
			findslot=-3;
	}
	return findslot;
}

void Node::insertEntry(long slot, Point& key, long id, NodePtr sibling, NodePtr downPtr)
{
	long split, i, j, k, x, y;
	if (sibling.get()==0)
	{
		placeEntry(slot+1, key, id, downPtr);
	}
	else
	{
		if (this->isIndex())
			i=1;
		else
			i=0;
		if (i==1)
			split=(m_capacity-m_minFanout)-1;
		else
			split=m_minFanout-1;
	        if (slot!=split)
			j=1;
		else
			j=0;
		if (slot>=split)
			k=1;
		else
			k=0;
		for (x=split+k+i*j, y=0; x<m_capacity; x++, y++)
		{
			sibling->m_ptrKey[y]=m_pTree->m_pointPool.acquire();
			*(sibling->m_ptrKey[y])=*(this->m_ptrKey[x]);
			sibling->m_pIdentifier[y]=this->m_pIdentifier[x];
			sibling->m_pDigest[y] = this->m_pDigest[x];
			this->m_children--;
			sibling->m_children++;
			if (this->m_embeddingLength>0)
			{	
				this->m_embeddingLength-=DIGEST_LENGTH;
				sibling->m_embeddingLength+=DIGEST_LENGTH;
			}
			for (int a=0;a<2;a++)
			{
				sibling->m_pID[y][a]=this->m_pID[x][a];
				sibling->m_pHash[y][a]=this->m_pHash[x][a];
				if (sibling->m_pHash[y][a]!=0)
				{
					if (this->m_embeddingLength>0)
					{
						this->m_embeddingLength-=DIGEST_LENGTH;
						sibling->m_embeddingLength+=DIGEST_LENGTH;
					}
				}	
			}
		}	
	       if (i==1)
		{
			if (j==1)
			{
				sibling->m_pFirstNode=this->m_pIdentifier[split+k];
				sibling->m_pFirstDigest=this->m_pDigest[split+k];
				this->m_children--; //as soon the new entry will be inserted and replace this entry at split+k;
			        if (this->m_embeddingLength>0)
				{				
					sibling->m_embeddingLength+=DIGEST_LENGTH;
					this->m_embeddingLength-=DIGEST_LENGTH;	
				}
			}
			else
			{
				sibling->m_pFirstNode=downPtr->m_identifier;
				if (downPtr->m_embeddingLength>0)
				{
					byte* digest=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));	
					sibling->computeNodeDigests(downPtr, digest);
					sibling->m_pFirstDigest=digest;
					sibling->m_embeddingLength+=DIGEST_LENGTH;
				}
				else
					sibling->m_pFirstDigest=0;
			}
		}		
		if (j==1)
		{
			Point tmpKey=*(m_ptrKey[split+k]);
			if (this->isLeaf())
			{	m_pTree->m_tmpKey=*(m_ptrKey[split+k]);
				//cout<<"update m_tmpKey "<<m_pTree->m_tmpKey.m_pCoords[0]<<endl;
			}
			if (k==1)
			{
				sibling->placeEntry(slot-split-i, key, id, downPtr);
			}
			else
				this->placeEntry(slot+1, key, id, downPtr);
			if (this->isIndex())
			{
				m_pTree->m_tmpKey=tmpKey;
				//cout<<"update m_tmpKey "<<m_pTree->m_tmpKey.m_pCoords[0]<<endl;
			}
		}
		else
		{
			m_pTree->m_tmpKey=key;
			if (i!=1)
				sibling->placeEntry(0, key, id, downPtr);
		}
	}
}

void Node::placeEntry(long slot, Point& key, long id, NodePtr downPtr)
{
	long i,x;
	if (m_children>0)
	{
		this->m_ptrKey[m_children]=m_pTree->m_pointPool.acquire();
		for (x=this->m_children-1; x>=slot; x--)
		{
			this->m_pIdentifier[x+1]=this->m_pIdentifier[x];
			*(this->m_ptrKey[x+1])=*(this->m_ptrKey[x]);		
			m_pDigest[x+1] = m_pDigest[x];

			for (int a=0;a<2;a++)
			{
				m_pID[x+1][a]=m_pID[x][a];
				m_pHash[x+1][a]=m_pHash[x][a];
			}
		}
	}
	if (slot==0)
		this->m_ptrKey[slot]=m_pTree->m_pointPool.acquire();
	if (this->isIndex())
	{
		this->m_pIdentifier[slot]=downPtr->m_identifier;
		*(this->m_ptrKey[slot])=m_pTree->m_tmpKey;
		this->m_children++;
	}			
	else
	{
		this->m_pIdentifier[slot]=id;
		//this->m_ptrKey[slot]=m_pTree->m_pointPool.acquire();
		*(this->m_ptrKey[slot])=key;
		this->m_children++;
	}
}

void Node::computeNodeDigests(NodePtr node, byte* digest)
{
	if (node->m_embeddingLength>0)
	{

		if (node->m_embeddingTree==0)
		{
			node->buildEmbeddings();
		}
		memcpy(digest, node->m_embeddingTree->m_treeDigest, DIGEST_LENGTH);
	}
	else
		digest=0;
}

void Node::getkeydigArray(long** keys, long** pids, int** ids, byte**& digests)
{
	long i;
	if (this->m_children==0)
	{
		*keys=0;
		digests=0;
	}
	else
	{
		if (this->isIndex())
		{
			*keys = new long[this->m_children+1];	
			*pids = new long[this->m_children+1];
			*ids = new int[this->m_children+1];
			digests = new byte*[this->m_children+1];
			(*keys)[0]=(long)m_ptrKey[0]->m_pCoords[0]-1;
			
			digests[0]=m_pFirstDigest;
			(*ids)[0]=m_children;
			(*pids)[0] = m_pFirstNode;
			for (i=0; i<m_children; i++)
			{
				(*keys)[i+1]=(long)m_ptrKey[i]->m_pCoords[0];
				digests[i+1]=m_pDigest[i];
				(*ids)[i+1]=i;
				(*pids)[i+1]=m_pIdentifier[i];
			}
		}
		else
		{
			*keys = new long[this->m_children];
			*pids = new long[this->m_children];
			*ids = new int[this->m_children];
			digests = new byte*[this->m_children];
			for (i=0; i<m_children; i++)
			{
				(*keys)[i]=(long)m_ptrKey[i]->m_pCoords[0];
				digests[i]=m_pDigest[i];
				(*ids)[i]=i;
				(*pids)[i]=m_pIdentifier[i];
			}
		}
	}
}

void Node::buildEmbeddings()
{
	if (m_embeddingTree!=0)
		delete m_embeddingTree;
	long* keys;
	byte** digests;
	int* ids;
	long* pids;
	getkeydigArray(&keys, &pids, &ids, digests);
	m_embeddingTree=new RBTree();
	if (this->isIndex())	
		m_embeddingTree->bulkloadTree(keys, pids, ids, digests, m_children+1);
	else
		m_embeddingTree->bulkloadTree(keys, pids, ids, digests, m_children);
	m_pTree->m_cputime+=m_embeddingTree->m_cputime;
	m_pTree->m_hashComputed+=m_embeddingTree->m_hashComputed;
	m_rootID=m_embeddingTree->m_rootID;
	delete[] keys;
	delete[] digests;
	delete[] ids;
	delete[] pids;
}

void Node::removeEntry(long slot)
{
	long x;
	if (this->isIndex())
		m_pTree->deleteNode( (m_pTree->readNode(this->m_pIdentifier[slot])).get() );
	if (m_embeddingLength>0)
	{	
		m_embeddingLength -= DIGEST_LENGTH;
		for (int a=0;a<2;a++)
		{
			if (m_pHash[slot][a]!=0)
				m_embeddingLength -= DIGEST_LENGTH;
		}
	}
	if (m_pDigest[slot]!=0)
		delete[] m_pDigest[slot];
	for (int i=0;i<2;i++)
		if (m_pHash[slot][i]!=0)
			delete[] m_pHash[slot][i];
	
	for (x=slot; x<m_children-1; x++)
	{
		m_pIdentifier[x]=m_pIdentifier[x+1];
		*(m_ptrKey[x])=*(m_ptrKey[x+1]);
		m_pDigest[x] = m_pDigest[x+1];

		for (int a=0;a<2;a++)
		{
			m_pID[x][a]=m_pID[x+1][a];
			m_pHash[x][a]=m_pHash[x+1][a];
		}
	}
	this->m_children--;
	return;
}

void Node::inorderTraverse(RBNode* node, long& pos)
{
	if (node==NULL || node==0)
		return;
	inorderTraverse(node->left, pos);
	if (pos==-1)
	{
		pos++;
		m_pFirstNode=node->m_pID;	
		if (node->cDigest!=NULL)
		{	if (m_pFirstDigest==0 || m_pFirstDigest==NULL)
				m_pFirstDigest=(byte*)malloc(DIGEST_LENGTH);
			memcpy(m_pFirstDigest,node->cDigest,DIGEST_LENGTH);
			m_embeddingLength+=DIGEST_LENGTH;
		}
	}
	else
	{
		if (node->m_id!=pos)
			cout<<"this should be wrong! "<<node->m_id<<" "<<pos<<endl;
		m_pIdentifier[pos]=node->m_pID;
		m_ptrKey[pos]->m_pCoords[0]=node->key;
		m_pID[pos][0]=node->m_leftID;
		m_pID[pos][1]=node->m_rightID;
		if (node->lDigest!=NULL)
		{
			if (m_pHash[pos][0]==0 || m_pHash[pos][0]==NULL)
				m_pHash[pos][0]=(byte*)malloc(DIGEST_LENGTH);
			
			memcpy(m_pHash[pos][0],node->lDigest,DIGEST_LENGTH);
			m_embeddingLength+=DIGEST_LENGTH;
		}	
		if (node->rDigest!=NULL)
		{	if (m_pHash[pos][1]==0 || m_pHash[pos][1]==NULL)
				m_pHash[pos][1]=(byte*)malloc(DIGEST_LENGTH);
			memcpy(m_pHash[pos][1],node->rDigest,DIGEST_LENGTH);
			m_embeddingLength+=DIGEST_LENGTH;
		}
		if (node->cDigest!=NULL)
		{	if (m_pDigest[pos]==0 || m_pDigest[pos]==NULL)
				m_pDigest[pos]=(byte*)malloc(DIGEST_LENGTH);
			memcpy(m_pDigest[pos],node->cDigest,DIGEST_LENGTH);
			m_embeddingLength+=DIGEST_LENGTH;
		}
		pos++;
	}
	inorderTraverse(node->right, pos);
}
