// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include "../spatialindex/SpatialIndexImpl.h"

#include "Node.h"
#include "Leaf.h"
#include "Index.h"
#include "BulkLoader.h"
#include "BTree.h"

using std::stack;
using std::vector;
using std::priority_queue;
using namespace std;

BTree::Data::Data(Point& r, long id)
	: m_id(id), m_point(r)
{
}

BTree::Data::~Data()
{
}

BTree::Data* BTree::Data::clone() throw (Tools::NotSupportedException)
{
	return new Data(m_point, m_id);
}

long BTree::Data::getIdentifier() const
{
	return m_id;
}

void BTree::Data::getShape(IShape** out) const throw (std::exception)
{
	*out = new Point(m_point);
}

void BTree::Data::getData(unsigned long& len, byte** data) const
{

}

unsigned long BTree::Data::getByteArraySize()
{
	return sizeof(long) + sizeof(unsigned long);
}

unsigned long BTree::Data::loadFromByteArray(byte* const data)
{

	return sizeof(long) + sizeof(unsigned long);
}

void BTree::Data::storeToByteArray(unsigned long& len, byte** data)
{
}

SpatialIndex::ISpatialIndex* SpatialIndex::BTree::returnBTree(SpatialIndex::IStorageManager& sm, Tools::PropertySet& ps)
	throw (std::exception, Tools::Exception)
{
	SpatialIndex::ISpatialIndex* si = new SpatialIndex::BTree::BTree(sm, ps);
	return si;
}

SpatialIndex::ISpatialIndex* SpatialIndex::BTree::createNewBTree(
	SpatialIndex::IStorageManager& sm,
	double fillFactor,
	unsigned long indexCapacity,
	unsigned long leafCapacity,
	unsigned long dimension,
	long embeddingFanout,
	BTreeVariant bv,
	long& indexIdentifier)
	throw (std::exception, Tools::Exception)
{
	unsigned long indexMinCapacity = ((unsigned long)(ceil((double(indexCapacity+1))/2.0)))-1;
	unsigned long leafMinCapacity = (unsigned long)( ceil(((double)leafCapacity)/2.0) );
 

	Tools::Variant var;
	Tools::PropertySet ps;

	var.m_varType = Tools::VT_DOUBLE;
	var.m_val.dblVal = fillFactor;
	ps.setProperty("FillFactor", var);

	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = indexCapacity;
	ps.setProperty("IndexCapacity", var);

	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = leafCapacity;
	ps.setProperty("LeafCapacity", var);

	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = indexMinCapacity;
	ps.setProperty("IndexMinFanout", var);

	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = leafMinCapacity;
	ps.setProperty("LeafMinFanout", var);


	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = dimension;
	ps.setProperty("Dimension", var);

	var.m_varType = Tools::VT_LONG;
	var.m_val.lVal = embeddingFanout;
	ps.setProperty("EmbeddingFanout", var);


	var.m_varType = Tools::VT_LONG;
	var.m_val.lVal = bv;
	ps.setProperty("TreeVariant", var);

	ISpatialIndex* ret = returnBTree(sm, ps);

	var = ps.getProperty("IndexIdentifier");
	indexIdentifier = var.m_val.lVal;

	return ret;
}

SpatialIndex::ISpatialIndex* SpatialIndex::BTree::createAndBulkLoadNewBTree(
	BulkLoadMethod m,
	IDataStream& stream,
	SpatialIndex::IStorageManager& sm,
	double fillFactor,
	unsigned long indexCapacity,
	unsigned long leafCapacity,
	unsigned long dimension,
	long embeddingFanout,
	SpatialIndex::BTree::BTreeVariant bv,
	long& indexIdentifier)
	throw (std::exception, Tools::Exception)
{
	return NULL;
}

SpatialIndex::ISpatialIndex* SpatialIndex::BTree::loadBTree(IStorageManager& sm, long indexIdentifier)
	throw (std::exception, Tools::Exception)
{
	Tools::Variant var;
	Tools::PropertySet ps;

	var.m_varType = Tools::VT_LONG;
	var.m_val.lVal = indexIdentifier;
	ps.setProperty("IndexIdentifier", var);

	return returnBTree(sm, ps);
}

SpatialIndex::BTree::BTree::BTree(IStorageManager& sm, Tools::PropertySet& ps) :
	m_pStorageManager(&sm),
	m_rootID(StorageManager::NewPage),
	m_headerID(StorageManager::NewPage),
	m_treeVariant(BV_PLUS),
	m_embeddingFanout(2),
	m_fillFactor(0.7),
	m_indexCapacity(100),
	m_leafCapacity(100),
	m_indexMinFanout(50),
	m_leafMinFanout(50),
	m_dimension(1),
	m_pointPool(1000),
	m_indexPool(100),
	m_leafPool(100),
	m_cputime(0.0),
	m_hashComputed(0),
	m_avgFanout(0),
	m_totalNode(0)
{
#ifdef PTHREADS
	pthread_rwlock_init(&m_rwLock, NULL);
#else
	m_rwLock = false;
#endif
	EVP_MD_CTX c;
	EVP_MD_CTX_init(&c);
	Tools::Variant var = ps.getProperty("IndexIdentifier");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_LONG) throw Tools::IllegalArgumentException("BTree: Property IndexIdentifier must be Tools::VT_LONG");
		m_headerID = var.m_val.lVal;
		initOld(ps);
	}
	else
	{
		initNew(ps);
		var.m_varType = Tools::VT_LONG;
		var.m_val.lVal = m_headerID;
		ps.setProperty("IndexIdentifier", var);
	}
}

SpatialIndex::BTree::BTree::~BTree()
{
#ifdef PTHREADS
	pthread_rwlock_destroy(&m_rwLock);
#endif

	storeHeader();
}

//
// ISpatialIndex interface
//

void SpatialIndex::BTree::BTree::insertData(unsigned long len, const byte* pData, const IShape& shape, long id) throw (Tools::Exception, std::exception)
{
	if (shape.getDimension() != m_dimension) throw Tools::IllegalArgumentException("insertData: Shape has the wrong number of dimensions.");

#ifdef PTHREADS
	Tools::ExclusiveLock lock(&m_rwLock);
#else
	if (m_rwLock == false) m_rwLock = true;
	else throw Tools::ResourceLockedException("insertData: cannot acquire an exclusive lock");
#endif

	try
	{
		// convert the shape into a Point (B-Trees index keys only; i.e. a point).
		PointPtr key = m_pointPool.acquire();
		shape.getCenter(*key);
		//pData is not used, as the hash value is computed on the fly
		//cout<<"Insert "<<key->m_pCoords[0]<<endl;
		insert(*key, id);
#ifndef PTHREADS
		m_rwLock = false;
#endif
	}
	catch (...)
	{
#ifndef PTHREADS
		m_rwLock = false;
#endif
		throw;
	}
}

bool SpatialIndex::BTree::BTree::deleteData(const IShape& shape, long id) throw (Tools::Exception, std::exception)
{
	if (shape.getDimension() != m_dimension) throw Tools::IllegalArgumentException("deleteData: Shape has the wrong number of dimensions.");

#ifdef PTHREADS
	Tools::ExclusiveLock lock(&m_rwLock);
#else
	if (m_rwLock == false) m_rwLock = true;
	else throw Tools::ResourceLockedException("deleteData: cannot acquire an exclusive lock");
#endif

	try
	{
		PointPtr key = m_pointPool.acquire();
		shape.getCenter(*key);
		bool ret = deleteData_impl(*key);

#ifndef PTHREADS
		m_rwLock = false;
#endif

		return ret;
	}
	catch (...)
	{
#ifndef PTHREADS
		m_rwLock = false;
#endif
		throw;
	}
}

void SpatialIndex::BTree::BTree::containsWhatQuery(const IShape& query, IVisitor& v) throw (Tools::Exception, std::exception)
{
	if (query.getDimension() != m_dimension) throw Tools::IllegalArgumentException("containsWhatQuery: Shape has the wrong number of dimensions.");
	rangeQuery(RangeQuery, query, v);
}

void SpatialIndex::BTree::BTree::intersectsWithQuery(const IShape& query, IVisitor& v) throw (Tools::Exception, std::exception)
{
	if (query.getDimension() != m_dimension) throw Tools::IllegalArgumentException("intersectsWithQuery: Shape has the wrong number of dimensions.");
	rangeQuery(RangeQuery, query, v);
}

void SpatialIndex::BTree::BTree::pointLocationQuery(const Point& query, IVisitor& v) throw (Tools::Exception, std::exception)
{
	if (query.m_dimension != m_dimension) throw Tools::IllegalArgumentException("pointLocationQuery: Shape has the wrong number of dimensions.");
	rangeQuery(SingleSearchQuery, query, v);
}

void SpatialIndex::BTree::BTree::nearestNeighborQuery(long k, const IShape& query, IVisitor& v, INearestNeighborComparator& nnc) throw (Tools::Exception, std::exception)
{
	if (query.getDimension() != m_dimension) throw Tools::IllegalArgumentException("nearestNeighborQuery: Shape has the wrong number of dimensions.");
}

void SpatialIndex::BTree::BTree::nearestNeighborQuery(long k, const IShape& query, IVisitor& v) throw (Tools::Exception, std::exception)
{
	if (query.getDimension() != m_dimension) throw Tools::IllegalArgumentException("nearestNeighborQuery: Shape has the wrong number of dimensions.");
}

void SpatialIndex::BTree::BTree::selfJoinQuery(const IShape& query, IVisitor& v) throw (Tools::Exception, std::exception)
{
	if (query.getDimension() != m_dimension) throw Tools::IllegalArgumentException("selfJoinQuery: Shape has the wrong number of dimensions.");

}

void SpatialIndex::BTree::BTree::queryStrategy(IQueryStrategy& qs) throw (Tools::Exception, std::exception)
{
#ifdef PTHREADS
	Tools::SharedLock lock(&m_rwLock);
#else
	if (m_rwLock == false) m_rwLock = true;
	else throw Tools::ResourceLockedException("queryStrategy: cannot acquire a shared lock");
#endif

	long next = m_rootID;
	bool hasNext = true;

	try
	{
		while (hasNext)
		{
			NodePtr n = readNode(next);
			qs.getNextEntry(*n, next, hasNext);
		}

#ifndef PTHREADS
		m_rwLock = false;
#endif
	}
	catch (...)
	{
#ifndef PTHREADS
		m_rwLock = false;
#endif
		throw;
	}
}

void SpatialIndex::BTree::BTree::getIndexProperties(Tools::PropertySet& out) const
{
	Tools::Variant var;

	// dimension
	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = m_dimension;
	out.setProperty("Dimension", var);

	// index capacity
	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = m_indexCapacity;
	out.setProperty("IndexCapacity", var);

	// leaf capacity
	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = m_leafCapacity;
	out.setProperty("LeafCapacity", var);

	// index min fanout 
	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = m_indexMinFanout;
	out.setProperty("IndexMinFanout", var);

	// leaf min fanout 
	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = m_leafMinFanout;
	out.setProperty("LeafMinFanout", var);

	// embedding fanout 
	var.m_varType = Tools::VT_LONG;
	var.m_val.lVal = m_embeddingFanout;
	out.setProperty("EmbeddingFanout", var);


	// B-tree variant
	var.m_varType = Tools::VT_LONG;
	var.m_val.lVal = m_treeVariant;
	out.setProperty("TreeVariant", var);

	// fill factor
	var.m_varType = Tools::VT_DOUBLE;
	var.m_val.dblVal = m_fillFactor;
	out.setProperty("FillFactor", var);

	// index pool capacity
	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = m_indexPool.getCapacity();
	out.setProperty("IndexPoolCapacity", var);

	// leaf pool capacity
	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = m_leafPool.getCapacity();
	out.setProperty("LeafPoolCapacity", var);

	// point pool capacity
	var.m_varType = Tools::VT_ULONG;
	var.m_val.ulVal = m_pointPool.getCapacity();
	out.setProperty("PointPoolCapacity", var);
}

void SpatialIndex::BTree::BTree::addCommand(ICommand* pCommand, CommandType ct)
{
	switch (ct)
	{
		case CT_NODEREAD:
			m_readNodeCommands.push_back(Tools::SmartPointer<ICommand>(pCommand));
			break;
		case CT_NODEWRITE:
			m_writeNodeCommands.push_back(Tools::SmartPointer<ICommand>(pCommand));
			break;
		case CT_NODEDELETE:
			m_deleteNodeCommands.push_back(Tools::SmartPointer<ICommand>(pCommand));
			break;
	}
}

//This function still needs to be modified... Basically, to ensure the integrity of the range check. Feifei
bool SpatialIndex::BTree::BTree::isIndexValid() throw (Tools::Exception, std::exception)
{
	/*
	bool ret = true;
	stack<ValidateEntry> st;
	NodePtr root = readNode(m_rootID);

	if (root->m_level != m_stats.m_treeHeight - 1)
	{
		std::cerr << "Invalid tree height." << std::endl;
		return false;
	}

	std::map<unsigned long, unsigned long> nodesInLevel;
	nodesInLevel.insert(std::pair<long, long>(root->m_level, 1));

	ValidateEntry e(root->m_nodeRange, root);
	st.push(e);

	while (! st.empty())
	{
		e = st.top(); st.pop();

		LineSegment tmpRange;
		tmpRange = m_infiniteRange;

		for (unsigned long cDim = 0; cDim < tmpRange.m_dimension; cDim++)
		{
			tmpRange.m_pStartPoint[cDim] = std::numeric_limits<double>::max();
			tmpRange.m_pEndPoint[cDim] = -std::numeric_limits<double>::max();

			for (unsigned long cChild = 0; cChild < e.m_pNode->m_children; cChild++)
			{
				tmpRange.m_pStartPoint[cDim] = std::min(tmpRange.m_pStartPoint[cDim], e.m_pNode->m_ptrKey[cChild]->m_pCoords[cDim]);
				tmpRange.m_pEndPoint[cDim] = std::max(tmpRange.m_pEndPoint[cDim], e.m_pNode->m_ptrKey[cChild]->m_pCoords[cDim]);
			}
		}

		if (! (tmpRange == e.m_pNode->m_nodeRange))
		{
			std::cerr << "Invalid parent information." << std::endl;
			ret = false;
		}
		else if (! (tmpRange == e.m_parentRange))
		{
			std::cerr << "Error in parent." << std::endl;
			ret = false;
		}

		if (e.m_pNode->m_level != 0)
		{
			for (unsigned long cChild = 0; cChild < e.m_pNode->m_children; cChild++)
			{
				NodePtr ptrN = readNode(e.m_pNode->m_pIdentifier[cChild]);
				ValidateEntry tmpEntry(ptrN->m_nodeRange, ptrN);

				std::map<unsigned long, unsigned long>::iterator itNodes = nodesInLevel.find(tmpEntry.m_pNode->m_level);

				if (itNodes == nodesInLevel.end())
				{
					nodesInLevel.insert(std::pair<long, long>(tmpEntry.m_pNode->m_level, 1l));
				}
				else
				{
					nodesInLevel[tmpEntry.m_pNode->m_level] = nodesInLevel[tmpEntry.m_pNode->m_level] + 1;
				}

				st.push(tmpEntry);
			}
		}
	}

	unsigned long nodes = 0;
	for (unsigned long cLevel = 0; cLevel < m_stats.m_treeHeight; cLevel++)
	{
		if (nodesInLevel[cLevel] != m_stats.m_nodesInLevel[cLevel])
		{
			std::cerr << "Invalid nodesInLevel information." << std::endl;
			ret = false;
		}

		nodes += m_stats.m_nodesInLevel[cLevel];
	}

	if (nodes != m_stats.m_nodes)
	{
		std::cerr << "Invalid number of nodes information." << std::endl;
		ret = false;
	}

	return ret;*/
	return true;
}

void SpatialIndex::BTree::BTree::getStatistics(IStatistics** out) const throw (std::exception)
{
	*out = new Statistics(m_stats);
}

void SpatialIndex::BTree::BTree::initNew(Tools::PropertySet& ps)
{
	Tools::Variant var;

	// tree variant
	var = ps.getProperty("TreeVariant");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (
			var.m_varType != Tools::VT_LONG ||
			var.m_val.lVal != BV_PLUS) 
			throw Tools::IllegalArgumentException("initNew: Property TreeVariant must be Tools::VT_LONG and of BTreeVariant type");

		m_treeVariant = static_cast<BTreeVariant>(var.m_val.lVal);
	}

	//embedding fanout
	var = ps.getProperty("EmbeddingFanout");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (
			var.m_varType != Tools::VT_LONG ||
			var.m_val.lVal < 2 ) 
			throw Tools::IllegalArgumentException("initNew: Property EmbeddingFanout must be Tools::VT_LONG and minimum 2");
		m_embeddingFanout = var.m_val.lVal;
	}


	// fill factor
	var = ps.getProperty("FillFactor");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (
			var.m_varType != Tools::VT_DOUBLE ||
			(var.m_val.dblVal <= 0.0 || var.m_val.dblVal>1.0))
			throw Tools::IllegalArgumentException("initNew: Property FillFactor must be Tools::VT_DOUBLE and in (0.0, 1.0)");
		m_fillFactor = var.m_val.dblVal;
	}

	// index capacity
	var = ps.getProperty("IndexCapacity");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG || var.m_val.ulVal < 1)
			throw Tools::IllegalArgumentException("initNew: Property IndexCapacity must be Tools::VT_ULONG and >= 1");

		m_indexCapacity = var.m_val.ulVal;
	}

	// leaf capacity
	var = ps.getProperty("LeafCapacity");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG || var.m_val.ulVal < 1)
			throw Tools::IllegalArgumentException("initNew: Property LeafCapacity must be Tools::VT_ULONG and >= 1");

		m_leafCapacity = var.m_val.ulVal;
	}
	
	//index min fanout
	var = ps.getProperty("IndexMinFanout");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG || var.m_val.ulVal < 1)
			throw Tools::IllegalArgumentException("initNew: Property IndexMinFanout must be Tools::VT_ULONG and >= 1");

		m_indexMinFanout = var.m_val.ulVal;
	}

	//leaf min fanout
	var = ps.getProperty("LeafMinFanout");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG || var.m_val.ulVal < 1)
			throw Tools::IllegalArgumentException("initNew: Property LeafMinFanout must be Tools::VT_ULONG and >= 1");

		m_leafMinFanout = var.m_val.ulVal;
	}

	// dimension
	var = ps.getProperty("Dimension");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG)
			throw Tools::IllegalArgumentException("initNew: Property Dimension must be Tools::VT_ULONG");
		if (var.m_val.ulVal < 1)
			throw Tools::IllegalArgumentException("initNew: Property Dimension must be at least 1");

		m_dimension = var.m_val.ulVal;
	}

	// index pool capacity
	var = ps.getProperty("IndexPoolCapacity");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG)
			throw Tools::IllegalArgumentException("initNew: Property IndexPoolCapacity must be Tools::VT_ULONG");

		m_indexPool.setCapacity(var.m_val.ulVal);
	}

	// leaf pool capacity
	var = ps.getProperty("LeafPoolCapacity");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG)
			throw Tools::IllegalArgumentException("initNew: Property LeafPoolCapacity must be Tools::VT_ULONG");

		m_leafPool.setCapacity(var.m_val.ulVal);
	}


	// point pool capacity
	var = ps.getProperty("PointPoolCapacity");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG)
			throw Tools::IllegalArgumentException("initNew: Property PointPoolCapacity must be Tools::VT_ULONG");

		m_pointPool.setCapacity(var.m_val.ulVal);
	}

	m_infinitePoint.makeInfinite(m_dimension);

	m_stats.m_treeHeight = 1;
	m_stats.m_nodesInLevel.push_back(0);

	Leaf root(this, -1);
	m_rootID = writeNode(&root);
	
	m_treeDigestLength=0;
	m_treeDigest=0;
	m_treeSignatureLength=0;
	m_treeSignature=0;

	storeHeader();
}

void SpatialIndex::BTree::BTree::initOld(Tools::PropertySet& ps)
{
	loadHeader();

	// only some of the properties may be changed.
	// the rest are just ignored.

	Tools::Variant var;

	// tree variant
	var = ps.getProperty("TreeVariant");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (
			var.m_varType != Tools::VT_LONG ||
			var.m_val.lVal != BV_PLUS) 
			throw Tools::IllegalArgumentException("initOld: Property TreeVariant must be Tools::VT_LONG and of BTreeVariant type");

		m_treeVariant = static_cast<BTreeVariant>(var.m_val.lVal);
	}

	//embedding fanout
	var = ps.getProperty("EmbeddingFanout");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (
			var.m_varType != Tools::VT_LONG ||
			var.m_val.lVal < 2 ) 
			throw Tools::IllegalArgumentException("initNew: Property EmbeddingFanout must be Tools::VT_LONG and minimum 2");
		m_embeddingFanout = var.m_val.lVal;
	}


	// index pool capacity
	var = ps.getProperty("IndexPoolCapacity");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG) throw Tools::IllegalArgumentException("initOld: Property IndexPoolCapacity must be Tools::VT_ULONG");

		m_indexPool.setCapacity(var.m_val.ulVal);
	}

	// leaf pool capacity
	var = ps.getProperty("LeafPoolCapacity");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG) throw Tools::IllegalArgumentException("initOld: Property LeafPoolCapacity must be Tools::VT_ULONG");

		m_leafPool.setCapacity(var.m_val.ulVal);
	}


	// point pool capacity
	var = ps.getProperty("PointPoolCapacity");
	if (var.m_varType != Tools::VT_EMPTY)
	{
		if (var.m_varType != Tools::VT_ULONG) throw Tools::IllegalArgumentException("initOld: Property PointPoolCapacity must be Tools::VT_ULONG");

		m_pointPool.setCapacity(var.m_val.ulVal);
	}

	m_infinitePoint.makeInfinite(m_dimension);
}

void SpatialIndex::BTree::BTree::storeHeader()
{
	const long headerSize =
		sizeof(long) +							// m_rootID
		sizeof(long) +							// m_treeVariant
		sizeof(long) + 							// m_embeddingFanout
		sizeof(double) +						// m_fillFactor
		sizeof(unsigned long) +					// m_indexCapacity
		sizeof(unsigned long) +					// m_leafCapacity
		sizeof(unsigned long) +					// m_dimension
		sizeof(unsigned long) +					// m_stats.m_nodes
		sizeof(unsigned long) +					// m_stats.m_data
		sizeof(unsigned long) +					// m_stats.m_treeHeight
		m_stats.m_treeHeight * sizeof(unsigned long) + // m_stats.m_nodesInLevel
		sizeof(unsigned long) +				//m_treeDigestLength
		m_treeDigestLength +				//m_treeDigest
		sizeof(unsigned long) +				//m_treeSignatureLength
		m_treeSignatureLength;				//m_treeSignature

	byte* header = new byte[headerSize];
	byte* ptr = header;

	memcpy(ptr, &m_rootID, sizeof(long));
	ptr += sizeof(long);
	memcpy(ptr, &m_treeVariant, sizeof(long));
	ptr += sizeof(long);
	memcpy(ptr, &m_embeddingFanout, sizeof(long));
	ptr += sizeof(long);
	memcpy(ptr, &m_fillFactor, sizeof(double));
	ptr += sizeof(double);
	memcpy(ptr, &m_indexCapacity, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	memcpy(ptr, &m_leafCapacity, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	memcpy(ptr, &m_dimension, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	memcpy(ptr, &(m_stats.m_nodes), sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	memcpy(ptr, &(m_stats.m_data), sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	memcpy(ptr, &(m_stats.m_treeHeight), sizeof(unsigned long));
	ptr += sizeof(unsigned long);

	for (unsigned long cLevel = 0; cLevel < m_stats.m_treeHeight; cLevel++)
	{
		memcpy(ptr, &(m_stats.m_nodesInLevel[cLevel]), sizeof(unsigned long));
		ptr += sizeof(unsigned long);
	}
	
	memcpy(ptr, &(m_treeDigestLength), sizeof(unsigned long));
	ptr+=sizeof(unsigned long);
	if (m_treeDigestLength>0)
	{
		memcpy(ptr, m_treeDigest, m_treeDigestLength);
		ptr+=m_treeDigestLength;
	}

	memcpy(ptr, &(m_treeSignatureLength), sizeof(unsigned long));
	ptr+=sizeof(unsigned long);
	if (m_treeSignatureLength>0)
	{
		memcpy(ptr, m_treeSignature, m_treeSignatureLength);
		ptr+=m_treeSignatureLength;
	}


	assert(headerSize == (ptr - header));

	m_pStorageManager->storeByteArray(m_headerID, headerSize, header);

	delete[] header;
}

void SpatialIndex::BTree::BTree::loadHeader()
{
	unsigned long headerSize;
	byte* header = 0;
	m_pStorageManager->loadByteArray(m_headerID, headerSize, &header);

	byte* ptr = header;

	memcpy(&m_rootID, ptr, sizeof(long));
	ptr += sizeof(long);
	memcpy(&m_treeVariant, ptr, sizeof(long));
	ptr += sizeof(long);
	memcpy(&m_embeddingFanout, ptr, sizeof(long));
	ptr += sizeof(long);
	memcpy(&m_fillFactor, ptr, sizeof(double));
	ptr += sizeof(double);
	memcpy(&m_indexCapacity, ptr, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	memcpy(&m_leafCapacity, ptr, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	memcpy(&m_dimension, ptr, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	memcpy(&(m_stats.m_nodes), ptr, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	memcpy(&(m_stats.m_data), ptr, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	memcpy(&(m_stats.m_treeHeight), ptr, sizeof(unsigned long));
	ptr += sizeof(unsigned long);

	m_indexMinFanout = ((unsigned long)(ceil((double(m_indexCapacity+1))/2.0)))-1;
	m_leafMinFanout = (unsigned long)( ceil(((double)m_leafCapacity)/2.0) );


	for (unsigned long cLevel = 0; cLevel < m_stats.m_treeHeight; cLevel++)
	{
		unsigned long cNodes;
		memcpy(&cNodes, ptr, sizeof(unsigned long));
		ptr += sizeof(unsigned long);
		m_stats.m_nodesInLevel.push_back(cNodes);
	}
	
	memcpy(&(m_treeDigestLength), ptr, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	if (m_treeDigestLength>0)
	{
		m_treeDigest=(byte*)malloc(m_treeDigestLength*sizeof(byte));
		memcpy(m_treeDigest, ptr, m_treeDigestLength);
		ptr += m_treeDigestLength;
	}	
	else
		m_treeDigest=0;

	memcpy(&(m_treeSignatureLength), ptr, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	if (m_treeSignatureLength>0)
	{
		m_treeSignature=(byte*)malloc(m_treeSignatureLength*sizeof(byte));
		memcpy(m_treeSignature, ptr, m_treeSignatureLength);
		ptr += m_treeSignatureLength;
	}	
	else
		m_treeSignature=0;

	delete[] header;
}

long SpatialIndex::BTree::BTree::compareKeys(long key1, long key2)
{
	long tmp=key1-key2;
	if (fabs(tmp)==0)
		tmp=0;
	return tmp;
}

SpatialIndex::BTree::NodePtr SpatialIndex::BTree::BTree::search(long searchKey, IVisitor& v)
{
	NodePtr findNode, root;
	root= readNode(m_rootID);
	findNode= descendToLeaf(searchKey, root, v);
	return findNode;
}


SpatialIndex::BTree::NodePtr SpatialIndex::BTree::BTree::descendToLeaf(long searchKey, NodePtr curr, IVisitor& v)
{
	long slot;
	NodePtr findNode;
	
	v.visitNode(*curr);
	for (slot=curr->getSlot(searchKey); curr->isIndex(); slot=curr->getSlot(searchKey))
	{
		byte** vos;
		long voSize;
		long* pos;				
		long posSize;
		if (curr->m_embeddingTree==0)
			curr->buildEmbeddings();
		curr->m_embeddingTree->constructVOs(searchKey, vos, voSize, &pos, posSize);
		v.updateVOs(vos, voSize, pos, posSize);
		if (slot>=0)
			curr= readNode(curr->m_pIdentifier[slot]);
		else
			curr= readNode(curr->m_pFirstNode);
		v.visitNode(*curr);
	}
	//if ((slot>=0) && (this->compareKeys(searchKey, curr->m_ptrKey[slot]->m_pCoords[0])==0.0))
	if (slot>=0)
	{
		findNode= curr;
		byte** vos;
		long voSize;
		long* pos;				
		long posSize;
		if (curr->m_embeddingTree==0)
			curr->buildEmbeddings();
		curr->m_embeddingTree->constructVOs(searchKey, vos, voSize, &pos, posSize);
		v.updateVOs(vos,voSize,pos, posSize);
	}
	else
		findNode= NodePtr(NULL);
	return findNode;
} 

void SpatialIndex::BTree::BTree::startCountingAffectedNodes()
{
	m_countAffectedNodes=true;
}

void SpatialIndex::BTree::BTree::stopCountingAffectedNodes()
{
	m_countAffectedNodes=false;
}

void SpatialIndex::BTree::BTree::clearCountAffectedNodes()
{
	m_affectedNodes.clear();
}

void SpatialIndex::BTree::BTree::insert(Point& key,long id)
{
	NodePtr newNode, root, curr,noNode;
	

	root=readNode(m_rootID);
	curr=root;
	
	m_splitNodeID=-1;	
	m_digestUpdateNodeID=-1;
	newNode=NodePtr(NULL);
	this->m_splitNodeID=-1;
	noNode=NodePtr(NULL);
	newNode=this->descendSplit(key, id, curr, noNode);
	if (newNode.get()!=0)
	{
		if (newNode->m_identifier !=m_splitNodeID)
		{
			//root=readNode(m_rootID);
			this->makeNewRoot(this->m_tmpKey, root,newNode);
			if (m_countAffectedNodes==true)
				m_affectedNodes.insert(pair<long, long>((long)m_rootID, 2));

		}
	}	
	this->updateTreeDigestSig();	
}

void SpatialIndex::BTree::BTree::makeNewRoot(Point& key, NodePtr oldRoot, NodePtr newNode)
{
	NodePtr newRoot;
	newRoot = m_indexPool.acquire();
	if (newRoot.get() == 0)
	{
		 newRoot = NodePtr(new Index(this, -1, 0), &m_indexPool);
	}
	newRoot->m_identifier=-1;
	newRoot->m_pFirstNode=oldRoot->m_identifier;
	byte* digest1=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
	newRoot->computeNodeDigests(oldRoot,digest1);
	newRoot->m_pFirstDigest=digest1;	
	newRoot->m_embeddingLength+=DIGEST_LENGTH;

	newRoot->m_ptrKey[0]= newRoot->m_pTree->m_pointPool.acquire();
	*(newRoot->m_ptrKey[0])=key;
	newRoot->m_pIdentifier[0]=newNode->m_identifier;
	byte* digest2=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
	newRoot->computeNodeDigests(newNode,digest2);
	newRoot->m_pDigest[0]=digest2;	
	newRoot->m_embeddingLength+=DIGEST_LENGTH;


	newRoot->m_children=1;
 	newRoot->m_level=oldRoot->m_level+1;

	m_stats.m_nodesInLevel.push_back(0);
	writeNode(static_cast<Node*>(newRoot.get()));
	m_rootID=newRoot->m_identifier;
	m_stats.m_treeHeight++;
}

SpatialIndex::BTree::NodePtr SpatialIndex::BTree::BTree::descendSplit(Point& searchKey, long id,  NodePtr curr, NodePtr parent)
{
	NodePtr newMe, newNode;
	long slot;

	if (!curr->isFull())
		this->m_splitNodeID=-1;
	else if(this->m_splitNodeID==-1)
		this->m_splitNodeID=curr->m_identifier;
	
	slot= curr->getSlot((long)searchKey.m_pCoords[0]);
	//cout<<"find slot "<<slot<<endl;
	if (curr->isIndex())
	{
		NodePtr child;
		if (slot>=0)
			child= readNode(curr->m_pIdentifier[slot]);
		else
			child= readNode(curr->m_pFirstNode);
		newMe= descendSplit(searchKey,id, child, curr);
	}				
	else if ((slot>=0) && curr->m_children>0 && compareKeys((long)searchKey.m_pCoords[0], (long)curr->m_ptrKey[slot]->m_pCoords[0])==0)
	{
		//key already existed in the BTree leaf node, do nothing
		newMe= NodePtr(NULL);
		m_splitNodeID=-1;
	}
	else
	{
		newMe= curr; 
		m_stats.m_data++;
	}
	newNode = NodePtr(NULL);
	
	if (newMe.get()!=0)
	{
		if (this->m_splitNodeID!= -1)
			newNode = split(curr);	
		curr->insertEntry(slot, searchKey, id, newNode, newMe);
		if (m_countAffectedNodes==true)
			m_affectedNodes.insert(pair<long, long>((long)curr->m_identifier, (long)curr->m_children));
		writeNode(static_cast<Node*>(curr.get()));	
		if (newNode.get()!=0)
		{
			writeNode(static_cast<Node*>(newNode.get()));
			if (m_countAffectedNodes==true)
				m_affectedNodes.insert(pair<long, long>((long)newNode->m_identifier, (long)newNode->m_children));
		}
		this->m_digestUpdateNodeID=curr->m_identifier;
	}
	if (this->m_digestUpdateNodeID!=-1 && parent.get()!=0)
	{
		NodePtr updatedNode=readNode(this->m_digestUpdateNodeID);	
		long z=parent->getSlot((long)updatedNode->m_ptrKey[0]->m_pCoords[0]);
		if (z>=0)
			parent->computeNodeDigests(updatedNode, parent->m_pDigest[z]);
		else
			parent->computeNodeDigests(updatedNode,parent->m_pFirstDigest);
		
		if (z<parent->m_children-1)	
			parent->computeNodeDigests(readNode(parent->m_pIdentifier[z+1]),parent->m_pDigest[z+1]);

		writeNode(static_cast<Node*>(parent.get()));
		this->m_digestUpdateNodeID=parent->m_identifier;
	}
	return newNode;
}

SpatialIndex::BTree::NodePtr SpatialIndex::BTree::BTree::split(NodePtr curr)
{
	NodePtr sibling;
	
	if (curr->isIndex()) sibling = m_indexPool.acquire();
	else if (curr->isLeaf()) sibling = m_leafPool.acquire();
	if (sibling.get() == 0)
	{	if (curr->isIndex())
	 		sibling = NodePtr(new Index(this, -1, curr->m_level), &m_indexPool);
		else
			sibling = NodePtr(new Leaf(this, -1), &m_leafPool);
	}
	sibling->m_level= curr->m_level;
	sibling->m_identifier=-1;
	sibling->m_pFirstDigest=0;
	sibling->m_embeddingTree=0;
	sibling->m_embeddingLength=0;
	memset(sibling->m_pDigest, 0, sibling->m_capacity*sizeof(byte*));
	writeNode(sibling.get());
	if (curr->isLeaf())
	{
		sibling->m_pFirstNode=curr->m_pFirstNode;
		curr->m_pFirstNode=sibling->m_identifier;
	}
	//writeNode(sibling);
	//writeNode(curr);
	if (m_splitNodeID == curr->m_identifier)
		m_splitNodeID= -1;
	m_stats.m_splits++;
	return sibling;
}

bool SpatialIndex::BTree::BTree::deleteData_impl(const Point& key)
{
	assert(key.m_dimension == m_dimension);
	NodePtr newNode;
	NodePtr nonNode=NodePtr(NULL); 

	NodePtr root = readNode(m_rootID);
	m_mergeNodeID=-1;
	m_digestUpdateNodeID=-1;
	newNode=descendBalance(key, root,nonNode, nonNode, nonNode, nonNode, nonNode); 
	if (newNode.get()!=0)
		collapseRoot(readNode(m_rootID), newNode);
	this->updateTreeDigestSig();
	return true;
}

void SpatialIndex::BTree::BTree::collapseRoot(NodePtr oldRoot, NodePtr newRoot)
{
	m_rootID=newRoot->m_identifier;
	m_stats.m_treeHeight--;
	deleteNode(oldRoot.get());
	writeNode(newRoot.get());
}

SpatialIndex::BTree::NodePtr SpatialIndex::BTree::BTree::descendBalance(const Point& key, NodePtr curr, NodePtr left, NodePtr right, NodePtr lanc, NodePtr ranc, NodePtr parent)
{
	NodePtr newMe, myLeft, myRight, lAnchor, rAnchor, newNode;
	bool notleft, notright, fewleft, fewright, test;
	int slot;

	if (!curr->isFewest())	
		m_mergeNodeID=-1;
	else if (m_mergeNodeID==-1)
		m_mergeNodeID=curr->m_identifier;

	slot = curr->getSlot((long)key.m_pCoords[0]);
	if (curr->isIndex())
	{
		if (slot>=0)
			newNode= readNode(curr->m_pIdentifier[slot]);
		else
			newNode= readNode(curr->m_pFirstNode);
		if (slot==-1)
		{
			if (left.get()==0)
				myLeft=NodePtr(NULL);
			else
				myLeft=readNode(left->m_pIdentifier[left->m_children-1]);
			lAnchor= lanc;
		}
		else
		{
				if (slot==0)
					myLeft=readNode(curr->m_pFirstNode);
				else
					myLeft= readNode(curr->m_pIdentifier[slot-1]);
				lAnchor =curr;
		}
		if (slot == curr->m_children-1)
		{
			if (right.get()==0)
				myRight=NodePtr(NULL);
			else
				myRight=readNode(right->m_pFirstNode);
			rAnchor= ranc;	
		}
		else
		{
			myRight=readNode(curr->m_pIdentifier[slot+1]);
			rAnchor= curr;
		}
		newMe = descendBalance(key, newNode, myLeft, myRight, lAnchor, rAnchor, curr);
	} 
	else if ( (slot>=0) && compareKeys((long)key.m_pCoords[0], (long)curr->m_ptrKey[slot]->m_pCoords[0])==0.0)
	{
		//cout<<"find deletion key in leaf node"<<endl;
		m_stats.m_data--;
		newMe= curr;
		newNode= curr;
	}
	else
	{
		newMe = NodePtr(NULL);
		m_mergeNodeID=-1;	
	}


	if (newMe.get()!=0)
	{	
		//cout<<"calling removeEntry "<<newMe->m_identifier<<" "<<newNode->m_identifier<<endl;
		//comment: should we delete newMe? It's no longer useful
		if (newMe->m_identifier!=newNode->m_identifier)
			curr->removeEntry(slot+1);
		else
			curr->removeEntry(slot);
		writeNode(curr.get());
	}
	
	if (m_mergeNodeID==-1)
		newNode=NodePtr(NULL);	
	else
	{
		//cout<<"there is node to be merged "<<m_mergeNodeID<<endl;
		notleft= (left.get()==0);
		notright= (right.get()==0);
		if (!notleft)
			fewleft= left->isFewest();
		else
			fewleft=true;
		if (!notright)
			fewright= right->isFewest();
		else 	
			fewright=true;

		if (notleft && notright)
		{
			//cout<<"case 1"<<endl;
			test= curr->isLeaf();
			if (test)
				newNode= NodePtr(NULL);
			else
				newNode= readNode(curr->m_pFirstNode);
		}	
		else if ((notleft || fewleft) && (notright || fewright))
		{
			//cout<<"case 2"<<endl;
			if (lanc.get()==0 && parent.get()==0)
				test=false;
			else if (lanc.get()==0)
				test=true;
			else if (parent.get()==0)
				test=true;
			else
				test = !(lanc->m_identifier == parent->m_identifier);
			if (test)
			{
				//cout<<"merge with right"<<endl;
				newNode= merge(curr, right, ranc);
			}
			else
			{
				//cout<<"merge with left"<<endl;
				newNode= merge(left, curr, lanc);
			}
		}
		else if (!notleft && fewleft && !notright && !fewright)
		{
			//cout<<"case 3"<<endl;
			if (ranc.get()!=0 && parent.get()!=0)
				test=(!(ranc->m_identifier== parent->m_identifier) && (curr->m_identifier==m_mergeNodeID));
			else if (ranc.get()==0 && parent.get()==0)
				test= (curr->m_identifier==m_mergeNodeID);
			else
				test=false;
			if (test)
				newNode= merge(left, curr, lanc);
			else
				newNode= shift(curr, right, ranc);
		}
		else if (!notleft && !fewleft && !notright && fewright)
		{
			//cout<<"case 4"<<endl;
			if (lanc.get()!=0 && parent.get()!=0)
				test=(!(lanc->m_identifier== parent->m_identifier) && (curr->m_identifier==m_mergeNodeID));
			else if (lanc.get()==0 && parent.get()==0)
				test= (curr->m_identifier==m_mergeNodeID);
			else
				test=false;		
	
			if (test)
				newNode= merge(curr, right, ranc);
			else
				newNode= shift(left, curr, lanc);
		}
		else if ( ((lanc.get()==0) && (ranc.get()==0)) || ( (lanc.get()!=0) && (ranc.get()!=0) && (lanc->m_identifier==ranc->m_identifier)))
		{
			//cout<<"case 5"<<endl;

			test=(left->m_children<right->m_children);
			if (test)
				newNode=shift(curr, right, ranc);
			else
				newNode=shift(left, curr, lanc);
		}
		else
		{
			//cout<<"case 6"<<endl;

			if (lanc.get()==0 && parent.get()==0)
				test=true;
			else if (lanc.get()==0)
				test=false;
			else if (parent.get()==0)
				test=false;
			else
				test= (lanc->m_identifier==parent->m_identifier);
			if (test)
				newNode=shift(left, curr, lanc);
			else
				newNode=shift(curr, right, ranc);
		}
	}
	//if (newNode.get()!=0)
		//writeNode(newNode.get());
	if (m_digestUpdateNodeID!=-1 && parent.get()!=0)
	{
		NodePtr updatedNode=readNode(m_digestUpdateNodeID);
		long z = parent->getSlot((long)updatedNode->m_ptrKey[0]->m_pCoords[0]);	
		if (z>=0)
			parent->computeNodeDigests(updatedNode, parent->m_pDigest[z]);
		else
			parent->computeNodeDigests(updatedNode, parent->m_pFirstDigest);
		m_digestUpdateNodeID=parent->m_identifier;
		writeNode(parent.get());
	}
	return newNode;
}

SpatialIndex::BTree::NodePtr SpatialIndex::BTree::BTree::shift( NodePtr left, NodePtr right, NodePtr anchor)
{
	long x, y, z;
	bool i=left->isIndex();
	
	//cout<<"calling shift"<<endl;
	if (left->m_children< right->m_children)
	{
		//cout<<"from right to left"<<endl;
		y= (right->m_children-left->m_children)>>1;
		x= left->m_children-1+y;
		m_tmpKey= *(right->m_ptrKey[y-i]);
		z=anchor->getSlot((long)m_tmpKey.m_pCoords[0]);
		if (i)
		{
			right->m_children--;
			left->m_children++;
			left->m_pIdentifier[left->m_children-1]=right->m_pFirstNode;
			left->m_ptrKey[left->m_children-1]=m_pointPool.acquire();
			*(left->m_ptrKey[left->m_children-1])=*(anchor->m_ptrKey[z]);

			left->m_pDigest[left->m_children-1]=right->m_pFirstDigest;
			left->m_embeddingLength+=DIGEST_LENGTH;
			right->m_embeddingLength-=DIGEST_LENGTH;

		}
		*(anchor->m_ptrKey[z])=m_tmpKey;
		for (z=y, y=y-1-i; y>=0; y--, x--)
		{
			right->m_children--;
			left->m_children++;
			left->m_pIdentifier[x]=right->m_pIdentifier[y];
			left->m_ptrKey[x]=m_pointPool.acquire();
			*(left->m_ptrKey[x])=*(right->m_ptrKey[y]);	
			
			left->m_pDigest[x]=right->m_pDigest[y];
			left->m_embeddingLength+=DIGEST_LENGTH;
			right->m_embeddingLength-=DIGEST_LENGTH;
		}
		for (x=0; x<right->m_children; x++)
		{
			right->m_pIdentifier[x]=right->m_pIdentifier[x+z];
			*(right->m_ptrKey[x])=*(right->m_ptrKey[x+z]);
			right->m_pDigest[x]=right->m_pDigest[x+z];
		}
		//here we assume the same size digest for old digest and new digest, if they are different in length, change this part! 
		if (z>0)
		{
			anchor->computeNodeDigests(left,anchor->m_pDigest[z-1]);		
			anchor->computeNodeDigests(right,anchor->m_pDigest[z]);
			
		}
		else
		{
			anchor->computeNodeDigests(left,anchor->m_pFirstDigest);	
			anchor->computeNodeDigests(right,anchor->m_pDigest[z]);
		}
	}
	else
	{
		//cout<<"from left to right"<<endl;
		y=(left->m_children-right->m_children)>>1;
		x=left->m_children-y;
		for (z=right->m_children-1;z>=0;z--)
		{
			if ((z+y)>right->m_children-1)
				right->m_ptrKey[z+y]=m_pointPool.acquire();
			*(right->m_ptrKey[z+y])=*(right->m_ptrKey[z]);
			right->m_pIdentifier[z+y]=right->m_pIdentifier[z];
			right->m_pDigest[z+y]=right->m_pDigest[z];
		}
		m_tmpKey=*(left->m_ptrKey[x]);
		z=anchor->getSlot((long)m_tmpKey.m_pCoords[0])+1;
		if (i)
		{
			left->m_children--;
			right->m_children++;
			right->m_pIdentifier[y-1]=right->m_pFirstNode;
			*(right->m_ptrKey[y-1])=*(anchor->m_ptrKey[z]);	
			right->m_pFirstNode=left->m_pIdentifier[x];	

			right->m_pFirstDigest=left->m_pDigest[x];
			right->m_embeddingLength+=DIGEST_LENGTH;
			left->m_embeddingLength-=DIGEST_LENGTH;
		}
		*(anchor->m_ptrKey[z])=m_tmpKey;
		for (x=left->m_children-1+i, y=y-1-i; y>=0; y--, x--)
		{
			left->m_children--;
			right->m_children++;
			right->m_pIdentifier[y]=left->m_pIdentifier[x];
			//right->m_ptrKey[y]=m_pointPool.acquire();
			*(right->m_ptrKey[y])=*(left->m_ptrKey[x]);	
			right->m_pDigest[y]=left->m_pDigest[x];
			right->m_embeddingLength+=DIGEST_LENGTH;
			left->m_embeddingLength-=DIGEST_LENGTH;
		}
		//here we assume the same size digest for old digest and new digest, if they are different in length, change this part! 
		if (z>0)
		{
			anchor->computeNodeDigests(left,anchor->m_pDigest[z-1]);		
			anchor->computeNodeDigests(right,anchor->m_pDigest[z]);
			
		}
		else
		{
			anchor->computeNodeDigests(left,anchor->m_pFirstDigest);	
			anchor->computeNodeDigests(right,anchor->m_pDigest[z]);
		}

	}
	m_stats.m_shift++;
	writeNode(left.get());
	writeNode(right.get());
	writeNode(anchor.get());
	m_mergeNodeID=-1;
	m_digestUpdateNodeID=anchor->m_identifier;
	
	return NodePtr(NULL);
}

SpatialIndex::BTree::NodePtr SpatialIndex::BTree::BTree::merge( NodePtr left, NodePtr right, NodePtr anchor)
{
	long x, y, z;
	
 	if (left->isIndex())
	{
		left->m_children++;
		m_tmpKey=*(right->m_ptrKey[0]);
		z= anchor->getSlot((long)m_tmpKey.m_pCoords[0]);
		m_tmpKey=*(anchor->m_ptrKey[z]);
		left->m_ptrKey[left->m_children-1]=m_pointPool.acquire();
		left->m_pIdentifier[left->m_children-1]=right->m_pFirstNode;
		*(left->m_ptrKey[left->m_children-1])=m_tmpKey;

		left->m_pDigest[left->m_children-1]=right->m_pFirstDigest;
		left->m_embeddingLength+=DIGEST_LENGTH;
		right->m_embeddingLength-=DIGEST_LENGTH;
	}	
	else
		left->m_pFirstNode=right->m_pFirstNode;

	for (x=left->m_children,y=0;y<right->m_children;x++,y++)
	{
		left->m_children++;
		left->m_pIdentifier[x]=right->m_pIdentifier[y];
		left->m_ptrKey[x]=m_pointPool.acquire();
		*(left->m_ptrKey[x])=*(right->m_ptrKey[y]);

		left->m_pDigest[x]=right->m_pDigest[y];
		left->m_embeddingLength+=DIGEST_LENGTH;
		right->m_embeddingLength-=DIGEST_LENGTH;
	}

	if (m_mergeNodeID==left->m_identifier || m_mergeNodeID==right->m_identifier)
		m_mergeNodeID=-1;
	m_stats.m_merge++;
	writeNode(left.get());
	//here we assume the same size digest for old digest and new digest, if they are different in length, change this part! 
	if (z>0)
		anchor->computeNodeDigests(left,anchor->m_pDigest[z-1]);		
	else 
		anchor->computeNodeDigests(left,anchor->m_pFirstDigest);	
	m_digestUpdateNodeID=anchor->m_identifier;
	return right;
}


void SpatialIndex::BTree::BTree::updateTreeDigestSig()
{
	NodePtr root=readNode(this->m_rootID);
	this->m_treeDigestLength=DIGEST_LENGTH;
	if (this->m_treeDigest==0)
		this->m_treeDigest=(byte*)malloc(this->m_treeDigestLength*sizeof(byte));	
	root->computeNodeDigests(root,this->m_treeDigest);
}


long SpatialIndex::BTree::BTree::writeNode(Node* n)
{
	byte* buffer;
	unsigned long dataLength;
	n->storeToByteArray(dataLength, &buffer);

	long page;
	if (n->m_identifier < 0) page = StorageManager::NewPage;
	else page = n->m_identifier;

	try
	{
		m_pStorageManager->storeByteArray(page, dataLength, buffer);
		delete[] buffer;
	}
	catch (Tools::InvalidPageException& e)
	{
		delete[] buffer;
		std::cerr << e.what() << std::endl;
		//std::cerr << *this << std::endl;
		throw Tools::IllegalStateException("writeNode: failed with Tools::InvalidPageException");
	}

	if (n->m_identifier < 0)
	{
		n->m_identifier = page;
		m_stats.m_nodes++;

#ifndef NDEBUG
		try
		{
			m_stats.m_nodesInLevel[n->m_level] = m_stats.m_nodesInLevel.at(n->m_level) + 1;
		}
		catch(...)
		{
			throw Tools::IllegalStateException("writeNode: writing past the end of m_nodesInLevel.");
		}
#else
		m_stats.m_nodesInLevel[n->m_level] = m_stats.m_nodesInLevel[n->m_level] + 1;
#endif
	}

	m_stats.m_writes++;

	for (unsigned long cIndex = 0; cIndex < m_writeNodeCommands.size(); cIndex++)
	{
		m_writeNodeCommands[cIndex]->execute(*n);
	}

	return page;
}

SpatialIndex::BTree::NodePtr SpatialIndex::BTree::BTree::readNode(unsigned long id)
{
	unsigned long dataLength;
	byte* buffer;

	try
	{
		m_pStorageManager->loadByteArray(id, dataLength, &buffer);
	}
	catch (Tools::InvalidPageException& e)
	{
		std::cerr << e.what() << std::endl;
		//std::cerr << *this << std::endl;
		throw Tools::IllegalStateException("readNode: failed with Tools::InvalidPageException");
	}

	try
	{
		long nodeType;
		memcpy(&nodeType, buffer, sizeof(long));

		NodePtr n;

		if (nodeType == PersistentIndex) n = m_indexPool.acquire();
		else if (nodeType == PersistentLeaf) n = m_leafPool.acquire();
		else throw Tools::IllegalStateException("readNode: failed reading the correct node type information");

		if (n.get() == 0)
		{
			if (nodeType == PersistentIndex) n = NodePtr(new Index(this, -1, 0), &m_indexPool);
			else if (nodeType == PersistentLeaf) n = NodePtr(new Leaf(this, -1), &m_leafPool);
		}

		//n->m_pTree = this;
		n->m_identifier = id;
		n->loadFromByteArray(buffer);

		m_stats.m_reads++;

		for (unsigned long cIndex = 0; cIndex < m_readNodeCommands.size(); cIndex++)
		{
			m_readNodeCommands[cIndex]->execute(*n);
		}

		delete[] buffer;
		return n;
	}
	catch (...)
	{
		delete[] buffer;
		throw;
	}
}

void SpatialIndex::BTree::BTree::deleteNode(Node* n)
{
	try
	{
		m_pStorageManager->deleteByteArray(n->m_identifier);
	}
	catch (Tools::InvalidPageException& e)
	{
		std::cerr << e.what() << std::endl;
		//std::cerr << *this << std::endl;
		throw Tools::IllegalStateException("deleteNode: failed with Tools::InvalidPageException");
	}

	m_stats.m_nodes--;
	m_stats.m_nodesInLevel[n->m_level] = m_stats.m_nodesInLevel[n->m_level] - 1;

	for (unsigned long cIndex = 0; cIndex < m_deleteNodeCommands.size(); cIndex++)
	{
		m_deleteNodeCommands[cIndex]->execute(*n);
	}
}

void SpatialIndex::BTree::BTree::rangeQuery(RangeQueryType type, const IShape& query, IVisitor& v) throw (Tools::Exception, std::exception)
{
#ifdef PTHREADS
	Tools::SharedLock lock(&m_rwLock);
#else
	if (m_rwLock == false) m_rwLock = true;
	else throw Tools::ResourceLockedException("rangeQuery: cannot acquire a shared lock");
#endif

	try
	{
		if (type==SingleSearchQuery)
		{
			const Point* key = dynamic_cast<const Point*>(&query);
			bool found=false;
			NodePtr res = search((long)key->m_pCoords[0],v);
			if (res.get()==0)
				cout<<"find no leaf node: cannot find data with this key"<<endl;
			else
			{
				long cChild= res->getSlot((long)key->m_pCoords[0]);
				if (compareKeys((long)(*(res->m_ptrKey[cChild])).m_pCoords[0],(long)key->m_pCoords[0])==0)
				{		Data data=Data(*(res->m_ptrKey[cChild]), res->m_pIdentifier[cChild]);
						v.visitData(data);
						m_stats.m_queryResults++;
						found=true;
				}
				/*
				for (long cChild=0; cChild< res->m_children; cChild++)
				{
					if ((*(res->m_ptrKey[cChild])).m_pCoords[0]==key->m_pCoords[0])
					{
						Data data=Data(*(res->m_ptrKey[cChild]), res->m_pIdentifier[cChild]);
						v.visitData(data);
						m_stats.m_queryResults++;
						found=true;
						break;
					}
				}*/
				if (found==false)
				{
					cout<<"find no key in returned leaf: cannot find data with this key"<<endl;	
				}
			}
		}
		else if (type=RangeQuery)
		{
			const LineSegment* range = dynamic_cast<const LineSegment*>(&query);
			NodePtr res = search((long)range->m_pStartPoint[0],v);
			if (res.get()==0)
				cout<<"cannot find data with this key"<<endl;
			else
			{
				long cChild=0;
				bool end=false;
				/*
				for (cChild=0; cChild< res->m_children; cChild++)
				{
					if ((*(res->m_ptrKey[cChild])).m_pCoords[0]>=range->m_pStartPoint[0])
						break;
				}*/
				cChild=res->getSlot((long)range->m_pStartPoint[0]);	
				long tmpComp=0;
			        while(!end)
				{		
					for (;cChild<res->m_children; cChild++)
					{
						
						tmpComp=compareKeys((long)range->m_pEndPoint[0], (long)(*(res->m_ptrKey[cChild])).m_pCoords[0]);
						if (tmpComp<0)
						{
							end=true;
							break;
						}
						Data data=Data(*(res->m_ptrKey[cChild]), res->m_pIdentifier[cChild]);
						v.visitData(data);
						m_stats.m_queryResults++;
						if (tmpComp==0)
						{
							end=true;
							break;
						}
					}
					if (!end)
					{
						if (res->m_pFirstNode!=-1)
						{	res = readNode(res->m_pFirstNode);
							v.visitNode(*res);
							cChild=0;	
						}
						else
							end=true;
					}
				}
			}
		}
#ifndef PTHREADS
		m_rwLock = false;
#endif
	}
	catch (...)
	{
#ifndef PTHREADS
		m_rwLock = false;
#endif
		throw;
	}
}

void SpatialIndex::BTree::BTree::showBTree(NodePtr root)
{
	std::vector<NodePtr> nodes;
	nodes.push_back(root);
	NodePtr tmp;
	while(!nodes.empty())
	{
		tmp=nodes[0];
		nodes.erase(nodes.begin());
		tmp->showNode();
		if (tmp->m_level!=0)
		{
			nodes.push_back(readNode(tmp->m_pFirstNode));	
			for (long cChild=0;cChild<tmp->m_children;cChild++)
				nodes.push_back(readNode(tmp->m_pIdentifier[cChild]));	
		}
	}
}

void SpatialIndex::BTree::BTree::showTree()
{
	NodePtr root=readNode(m_rootID);
	showBTree(root);
}

void SpatialIndex::BTree::BTree::computeFanout()
{
	NodePtr root=readNode(m_rootID);
	m_avgFanout=0;
	m_totalNode=0;

	std::vector<NodePtr> nodes;
	nodes.push_back(root);
	NodePtr tmp;
	while(!nodes.empty())
	{
		tmp=nodes[0];
		nodes.erase(nodes.begin());
		m_avgFanout+=tmp->m_children;
		m_totalNode++;
		if (tmp->m_level!=0)
		{
			nodes.push_back(readNode(tmp->m_pFirstNode));	
			for (long cChild=0;cChild<tmp->m_children;cChild++)
				nodes.push_back(readNode(tmp->m_pIdentifier[cChild]));	
		}
	}
	m_avgFanout=(unsigned long)((double)m_avgFanout/(double)m_totalNode);
}

std::ostream& SpatialIndex::BTree::operator<<(std::ostream& os, const BTree& t)
{
	using std::endl;

	os	<< "Dimension: " << t.m_dimension << endl
		<< "Fill factor: " << t.m_fillFactor << endl
		<< "Embedding Fanout: " << t.m_embeddingFanout << endl
		<< "Index capacity: " << t.m_indexCapacity << endl
		<< "Leaf capacity: " << t.m_leafCapacity << endl
		<< "Index MinFanout: " << t.m_indexMinFanout << endl
		<< "Leaf MinFanout: " << t.m_leafMinFanout << endl;



	if (t.m_stats.getNumberOfNodesInLevel(0) > 0)
		os	<< "Utilization: " << 100 * t.m_stats.getNumberOfData() / (t.m_stats.getNumberOfNodesInLevel(0) * t.m_leafCapacity) << "%" << endl
			<< t.m_stats;

	#ifndef NDEBUG
	os	<< "Leaf pool hits: " << t.m_leafPool.m_hits << endl
		<< "Leaf pool misses: " << t.m_leafPool.m_misses << endl
		<< "Index pool hits: " << t.m_indexPool.m_hits << endl
		<< "Index pool misses: " << t.m_indexPool.m_misses << endl
		<< "Point pool hits: " << t.m_pointPool.m_hits << endl
		<< "Point pool misses: " << t.m_pointPool.m_misses << endl;
	#endif

	os 	<< "Tree digest: ";
	char* d= printDigest(t.m_treeDigest, t.m_treeDigestLength); 
	for (unsigned long i=0; i< t.m_treeDigestLength*2; i++)
		os << d[i];
	os << endl;
	os	<< "Computation cost: "<<t.m_cputime <<endl;
	os	<< "Total hash computed: "<<t.m_hashComputed<<endl;
	os	<< "Total nodes: " <<t.m_totalNode <<endl;
	os	<< "Average fanout: "<<t.m_avgFanout <<endl;

	delete[] d;
	return os;
}
