// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

// NOTE: Please read README.txt before browsing this code.

// include library header file.
#include <SpatialIndex.h>

using namespace SpatialIndex;
using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2

// example of a Visitor pattern.
// findes the index and leaf IO for answering the query and prints
// the resulting data IDs to stdout.
class MyVisitor : public IVisitor
{
public:
	unsigned long m_indexIO;
	unsigned long m_leafIO;
	unsigned long m_total;

public:
	MyVisitor() : m_indexIO(0), m_leafIO(0), m_total(0) {}

	void visitNode(const INode& n)
	{
		if (n.isLeaf()) m_leafIO++;
		else m_indexIO++;
	}

	void visitData(const IData& d)
	{
		IShape* pS;
		d.getShape(&pS);
			// do something.
		delete pS;

		// data should be an array of characters representing a Region as a string.
		//byte* pData = 0;
		//unsigned long cLen = 0;
		//d.getData(cLen, &pData);
			// do something.
			//string s = reinterpret_cast<char*>(pData);
			//cout << s << endl;
		//delete[] pData;
		m_total++;

		//cout << d.getIdentifier() << endl;
			// the ID of this data entry is an answer to the query. I will just print it to stdout.
	}

	void visitData(std::vector<const IData*>& v)
	{
		cout << v[0]->getIdentifier() << " " << v[1]->getIdentifier() << endl;
	}
};

int main(int argc, char** argv)
{
	try
	{
		if (argc != 4)
		{
			cerr << "Usage: " << argv[0] << " query_file tree_file query_type [pointsearch | rangequery]." << endl;
			return -1;
		}

		int queryType = 0;

		if (strcmp(argv[3], "pointsearch") == 0) queryType = 0;
		else if (strcmp(argv[3], "rangequery") == 0) queryType = 1;
		else
		{
			cerr << "Unknown query type." << endl;
			return -1;
		}

		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open query file " << argv[1] << "." << endl;
			return -1;
		}

		string baseName = argv[2];
		IStorageManager* diskfile = StorageManager::loadDiskStorageManager(baseName);
			// this will try to locate and open an already existing storage manager.

		StorageManager::IBuffer* file = StorageManager::createNewRandomEvictionsBuffer(*diskfile, 10, false);
			// applies a main memory random buffer on top of the persistent storage manager
			// (LRU buffer, etc can be created the same way).

		// If we need to open an existing tree stored in the storage manager, we only
		// have to specify the index identifier as follows
		ISpatialIndex* tree = BTree::loadBTree(*file, 1);

		unsigned long count = 0;
		unsigned long indexIO = 0;
		unsigned long leafIO = 0;
		unsigned long total=0;
		unsigned long minrange=100000;
		unsigned long maxrange=0;
		unsigned long id, op;
		double x[1], y[1];
		Tools::ResourceUsage ru;

		ru.start();

		while (fin)
		{
			fin >> op >> id >> x[0] >> y[0];
			if (! fin.good()) continue; // skip newlines, etc.

			if (op == QUERY)
			{

				MyVisitor vis;

				if (queryType == 0)
				{
					Point p = Point(x, 1);
					tree->pointLocationQuery(p, vis);
				}
				else if (queryType == 1)
				{
					LineSegment r = LineSegment(x, y, 1);
					tree->containsWhatQuery(r, vis);
				}
				indexIO += vis.m_indexIO;
				leafIO += vis.m_leafIO;
				total += vis.m_total;
				if (vis.m_total<minrange)
					minrange=vis.m_total;
				else if (vis.m_total>maxrange)
					maxrange=vis.m_total;
					// example of the Visitor pattern usage, for calculating how many nodes
					// were visited.
			}
			else
			{
				cerr << "This is not a query operation." << endl;
			}

			if ((count % 1000) == 0 && count!=0) cerr << count << endl;

			count++;
		}

		ru.stop();
		/*
		map<int, int> test;
		test.insert(pair<int, int>(1,3));
		test.insert(pair<int, int>(2,6));
		test.insert(pair<int, int>(1,5));
		cerr <<"map size: "<<test.size()<< endl;
		map<int, int>::iterator it = test.begin();
		for (it;it!=test.end();it++)
			cerr<<(*it).first<<" "<<(*it).second<<endl;*/

		
	
		cerr << "Operations: " << count << endl;
		cerr << *tree;
		cerr << "Index I/O: " << indexIO << endl;
		cerr << "Leaf I/O: " << leafIO << endl;
		cerr << "Total records: " << total <<endl;
		cerr << "Min range: " << minrange <<endl;
		cerr << "Max range: " << maxrange <<endl;
		cerr << "Buffer hits: " << file->getHits() << endl;
		cerr << "Seconds: " << ru.getTotalTime() << endl;

		delete tree;
		delete file;
		delete diskfile;
			// delete the buffer first, then the storage manager
			// (otherwise the the buffer will fail writting the dirty entries).
	}
	catch (Tools::Exception& e)
	{
		cerr << "******ERROR******" << endl;
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "******ERROR******" << endl;
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
